from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn.functional as F
import torch.utils.checkpoint
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers.utils import logging

logger = logging.get_logger(__name__)


class PointerHead(torch.nn.Module):
    # Copy and generate,
    def __init__(
        self,
        # (decoder) input space
        target_token_ids: List[int],
        # output space (targets)
        bos_id: int,
        eos_id: int,
        pad_id: int,
        # embeddings
        embeddings: nn.Embedding,
        embedding_weight_mapping: Optional[Dict[Union[int, str], List[int]]] = None,
        # other parameters
        use_encoder_mlp: bool = False,
        use_constraints_encoder_mlp: bool = False,
        decoder_position_id_mode: Optional[nn.Module] = None,
        decoder_position_id_pattern: Optional[List[int]] = None,
        decoder_position_id_mapping: Optional[Dict[str, int]] = None,
    ):
        super().__init__()

        self.embeddings = embeddings

        self.pointer_offset = len(target_token_ids)

        # check that bos, eos, and pad are not out of bounds
        for target_id, target_id_name in zip(
            [bos_id, eos_id, pad_id], ["bos_id", "eos_id", "pad_id"]
        ):
            if target_id >= len(target_token_ids):
                raise ValueError(
                    f"{target_id_name} [{target_id}] must be smaller than the number of target token ids "
                    f"[{len(target_token_ids)}]!"
                )

        self.bos_id = bos_id
        self.eos_id = eos_id
        self.pad_id = pad_id
        # all ids that are not bos, eos or pad are label ids
        self.label_ids = [
            target_id
            for target_id in range(len(target_token_ids))
            if target_id not in [self.bos_id, self.eos_id, self.pad_id]
        ]

        target2token_id = torch.LongTensor(target_token_ids)
        self.register_buffer("target2token_id", target2token_id)
        self.label_token_ids = self.target2token_id[self.label_ids]
        self.eos_token_id = target_token_ids[self.eos_id]
        self.pad_token_id = target_token_ids[self.pad_id]

        hidden_size = self.embeddings.embedding_dim
        if use_encoder_mlp:
            self.encoder_mlp = nn.Sequential(
                nn.Linear(hidden_size, hidden_size),
                nn.Dropout(0.3),
                nn.ReLU(),
                nn.Linear(hidden_size, hidden_size),
            )
        if use_constraints_encoder_mlp:
            self.constraints_encoder_mlp = nn.Sequential(
                nn.Linear(hidden_size, hidden_size),
                nn.Dropout(0.3),
                nn.ReLU(),
                nn.Linear(hidden_size, hidden_size),
            )

        self.embedding_weight_mapping = None
        if embedding_weight_mapping is not None:
            # Because of config serialization, the keys may be strings. Convert them back to ints.
            self.embedding_weight_mapping = {
                int(k): v for k, v in embedding_weight_mapping.items()
            }

        self.decoder_position_id_mode = decoder_position_id_mode
        self.decoder_position_id_mapping = decoder_position_id_mapping
        if self.decoder_position_id_mode is None:
            pass
        elif self.decoder_position_id_mode in ["pattern", "pattern_with_increment"]:
            if decoder_position_id_pattern is None:
                raise ValueError(
                    "decoder_position_id_pattern must be provided when using "
                    'decoder_position_id_mode="pattern" or "pattern_with_increment"!'
                )
            self.register_buffer(
                "decoder_position_id_pattern", torch.tensor(decoder_position_id_pattern)
            )
        elif self.decoder_position_id_mode == "mapping":
            if self.decoder_position_id_mapping is None:
                raise ValueError(
                    'decoder_position_id_mode="mapping" requires decoder_position_id_mapping to be provided!'
                )
        else:
            raise ValueError(
                f'decoder_position_id_mode="{self.decoder_position_id_mode}" is not supported, '
                'use one of "pattern", "pattern_with_increment", or "mapping"!'
            )

    @property
    def use_prepared_position_ids(self):
        return self.decoder_position_id_mode is not None

    def set_embeddings(self, embedding: nn.Embedding) -> None:
        self.embeddings = embedding

    def overwrite_embeddings_with_mapping(self) -> None:
        """Overwrite individual embeddings with embeddings for other tokens.

        This is useful, for instance, if the label vocabulary is a subset of the source vocabulary.
        In this case, this method can be used to initialize each label embedding with one or
        multiple (averaged) source embeddings.
        """
        if self.embedding_weight_mapping is not None:
            for special_token_index, source_indices in self.embedding_weight_mapping.items():
                self.embeddings.weight.data[special_token_index] = self.embeddings.weight.data[
                    source_indices
                ].mean(dim=0)

    def prepare_decoder_input_ids(
        self,
        input_ids: torch.LongTensor,
        encoder_input_ids: torch.LongTensor,
    ) -> torch.LongTensor:
        mapping_token_mask = input_ids.lt(self.pointer_offset)
        mapped_tokens = input_ids.masked_fill(input_ids.ge(self.pointer_offset), 0)
        tag_mapped_tokens = self.target2token_id[mapped_tokens]

        encoder_input_ids_index = input_ids - self.pointer_offset
        encoder_input_ids_index = encoder_input_ids_index.masked_fill(
            encoder_input_ids_index.lt(0), 0
        )
        encoder_input_length = encoder_input_ids.size(1)
        if encoder_input_ids_index.max() >= encoder_input_length:
            raise ValueError(
                f"encoder_input_ids_index.max() [{encoder_input_ids_index.max()}] must be smaller "
                f"than encoder_input_length [{encoder_input_length}]!"
            )

        word_mapped_tokens = encoder_input_ids.gather(index=encoder_input_ids_index, dim=1)

        decoder_input_ids = torch.where(
            mapping_token_mask, tag_mapped_tokens, word_mapped_tokens
        ).to(torch.long)
        assert isinstance(decoder_input_ids, torch.LongTensor)

        # Note: we do not need to explicitly handle the padding (via a decoder attention mask) because
        # it gets automatically mapped to the pad token id

        return decoder_input_ids

    def prepare_decoder_position_ids(self, input_ids: torch.LongTensor) -> torch.LongTensor:
        if self.decoder_position_id_mode in ["pattern", "pattern_with_increment"]:
            bsz, tokens_len = input_ids.size()
            pattern_len = len(self.decoder_position_id_pattern)
            # the number of full and partly records. note that tokens_len includes the bos token
            repeat_num = (tokens_len - 2) // pattern_len + 1
            position_ids = self.decoder_position_id_pattern.repeat(bsz, repeat_num)

            if self.decoder_position_id_mode == "pattern_with_increment":
                position_ids_reshaped = position_ids.view(bsz, -1, pattern_len)
                add_shift_pos = (
                    torch.arange(0, repeat_num, device=position_ids_reshaped.device)
                    .repeat(bsz)
                    .view(bsz, -1)
                    .unsqueeze(-1)
                )
                # multiply by the highest position id in the pattern so that the position ids are unique
                # for any decoder_position_id_pattern across all records
                add_shift_pos *= max(self.decoder_position_id_pattern) + 1
                position_ids_reshaped = add_shift_pos + position_ids_reshaped
                position_ids = position_ids_reshaped.view(bsz, -1).long()
            # use start_position_id=0
            start_pos = torch.zeros(bsz, 1, dtype=position_ids.dtype, device=position_ids.device)
            # shift by 2 to account for start_position_id=0 and pad_position_id=1
            all_position_ids = torch.cat([start_pos, position_ids + 2], dim=-1)
            all_position_ids_truncated = all_position_ids[:bsz, :tokens_len]

            # mask the padding tokens
            mask_invalid = input_ids.eq(self.pad_id)
            all_position_ids_truncated_masked = all_position_ids_truncated.masked_fill(
                mask_invalid, 1
            ).to(torch.long)
            assert isinstance(all_position_ids_truncated_masked, torch.LongTensor)
            return all_position_ids_truncated_masked
        elif self.decoder_position_id_mode == "mapping":
            # we ignor the typing issue here because we ensure that the mapping is not None in the __init__
            mapping: Dict[str, int] = self.decoder_position_id_mapping  # type: ignore
            if "default" not in mapping:
                raise ValueError(
                    f"mapping must contain a default entry, but only contains {list(mapping)}!"
                )
            position_ids = input_ids.new_full(input_ids.size(), fill_value=mapping["default"])
            # ensure that values for all vocab entries are set first
            if "vocab" in mapping:
                position_ids[input_ids.lt(self.pointer_offset)] = mapping["vocab"]
            already_set: Dict[int, Tuple[str, int]] = {}
            for key, value in mapping.items():
                if key in ["default", "vocab"]:
                    continue
                elif key == "bos":
                    input_id = self.bos_id
                elif key == "eos":
                    input_id = self.eos_id
                elif key == "pad":
                    input_id = self.pad_id
                else:
                    raise ValueError(f"Mapping contains unknown key '{key}' (mapping: {mapping}).")
                if already_set.get(input_id, (key, value))[1] != value:
                    previous_key, previous_value = already_set[input_id]
                    raise ValueError(
                        f"Can not set the position ids for '{key}' to {value} because it was already "
                        f"set to {previous_value} by key '{previous_key}'. Note that both, '{key}' and "
                        f"'{previous_key}', have the same id ({input_id}), so their position_ids need to "
                        f"be also the same (position id mapping: {mapping})."
                    )
                position_ids[input_ids.eq(input_id)] = value
                already_set[input_id] = key, value
            return position_ids
        else:
            raise ValueError(
                f"decoder_position_id_mode={self.decoder_position_id_mode} not supported!"
            )

    def prepare_decoder_inputs(
        self,
        input_ids: torch.LongTensor,
        encoder_input_ids: torch.LongTensor,
        position_ids: Optional[torch.LongTensor] = None,
    ) -> Dict[str, torch.Tensor]:
        inputs: Dict[str, torch.Tensor] = {}
        if self.use_prepared_position_ids:
            if position_ids is None:
                position_ids = self.prepare_decoder_position_ids(input_ids=input_ids)
            inputs["position_ids"] = position_ids

        inputs["input_ids"] = self.prepare_decoder_input_ids(
            input_ids=input_ids,
            encoder_input_ids=encoder_input_ids,
        )
        return inputs

    def forward(
        self,
        last_hidden_state,
        encoder_input_ids,
        encoder_last_hidden_state,
        encoder_attention_mask,
        labels: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        constraints: Optional[torch.LongTensor] = None,
    ):
        # assemble the logits
        logits = last_hidden_state.new_full(
            (
                last_hidden_state.size(0),
                last_hidden_state.size(1),
                self.pointer_offset + encoder_input_ids.size(-1),
            ),
            fill_value=-1e24,
        )

        # eos and label scores depend only on the decoder output
        # bsz x max_len x 1
        eos_scores = F.linear(last_hidden_state, self.embeddings.weight[[self.eos_token_id]])
        label_embeddings = self.embeddings.weight[self.label_token_ids]
        # bsz x max_len x num_class
        label_scores = F.linear(last_hidden_state, label_embeddings)

        # the pointer depends on the src token embeddings, the encoder output and the decoder output
        # bsz x max_bpe_len x hidden_size
        src_outputs = encoder_last_hidden_state
        if getattr(self, "encoder_mlp", None) is not None:
            src_outputs = self.encoder_mlp(src_outputs)

        # bsz x max_word_len x hidden_size
        input_embed = self.embeddings(encoder_input_ids)

        # bsz x max_len x max_word_len
        word_scores = torch.einsum("blh,bnh->bln", last_hidden_state, src_outputs)
        gen_scores = torch.einsum("blh,bnh->bln", last_hidden_state, input_embed)
        avg_word_scores = (gen_scores + word_scores) / 2

        # never point to the padding or the eos token in the encoder input
        # TODO: why not excluding the bos token? seems to give worse results, but not tested extensively
        mask_invalid = encoder_attention_mask.eq(0) | encoder_input_ids.eq(self.eos_token_id)
        avg_word_scores = avg_word_scores.masked_fill(mask_invalid.unsqueeze(1), -1e32)

        # Note: the remaining row in logits contains the score for the bos token which should be never generated!
        logits[:, :, [self.eos_id]] = eos_scores
        logits[:, :, self.label_ids] = label_scores
        logits[:, :, self.pointer_offset :] = avg_word_scores

        loss = None
        # compute the loss if labels are provided
        if labels is not None:
            loss_fct = CrossEntropyLoss()
            logits_resized = logits.reshape(-1, logits.size(-1))
            labels_resized = labels.reshape(-1)
            if decoder_attention_mask is None:
                raise ValueError("decoder_attention_mask must be provided to compute the loss!")
            mask_resized = decoder_attention_mask.reshape(-1)
            labels_masked = labels_resized.masked_fill(
                ~mask_resized.to(torch.bool), loss_fct.ignore_index
            )
            loss = loss_fct(logits_resized, labels_masked)

        # compute the constraints loss if constraints are provided
        if constraints is not None:
            if getattr(self, "constraints_encoder_mlp", None) is not None:
                # TODO: is it fine to apply constraints_encoder_mlp to both src_outputs and label_embeddings?
                #  This is what the original code seems to do, but this is different from the usage of encoder_mlp.
                constraints_src_outputs = self.constraints_encoder_mlp(src_outputs)
                constraints_label_embeddings = self.constraints_encoder_mlp(label_embeddings)
            else:
                constraints_src_outputs = src_outputs
                constraints_label_embeddings = label_embeddings
            constraints_label_scores = F.linear(last_hidden_state, constraints_label_embeddings)
            # bsz x max_len x max_word_len
            constraints_word_scores = torch.einsum(
                "blh,bnh->bln", last_hidden_state, constraints_src_outputs
            )
            constraints_logits = last_hidden_state.new_full(
                (
                    last_hidden_state.size(0),
                    last_hidden_state.size(1),
                    self.pointer_offset + encoder_input_ids.size(-1),
                ),
                fill_value=-1e24,
            )
            constraints_logits[:, :, self.label_ids] = constraints_label_scores
            constraints_logits[:, :, self.pointer_offset :] = constraints_word_scores

            mask = constraints >= 0
            constraints_logits_valid = constraints_logits[mask]
            constraints_valid = constraints[mask]
            loss_c = F.binary_cross_entropy(
                torch.sigmoid(constraints_logits_valid), constraints_valid.float()
            )

            if loss is None:
                loss = loss_c
            else:
                loss += loss_c

        return logits, loss



import logging
from typing import Any, Callable, Dict, Generic, Optional, Sequence, Tuple, TypeVar

import torch
from torch.nn import ModuleDict
from torchmetrics import Metric

from .common import MetricWithArbitraryCounts

logger = logging.getLogger(__name__)
T = TypeVar("T")
U = TypeVar("U")


class WrappedLayerMetricsWithUnbatchAndDecodeWithErrorsFunction(
    MetricWithArbitraryCounts, Generic[T, U]
):
    """A wrapper around annotation layer metrics that can be used with batched encoded annotations.

    Args:
        layer_metrics: A dictionary mapping layer names to annotation layer metrics. Each metric
            should be a subclass of torchmetrics.Metric and should take two sets of annotations as
            input.
        unbatch_function: A function that takes a batched input and returns an iterable of
            individual inputs. This is used to unbatch the input before passing it to the annotation
            decoding function (decode_annotations_with_errors_function).
        decode_layers_with_errors_function: A function that takes an annotation encoding and
            returns a tuple of two dictionaries. The first dictionary maps layer names to a list of
            annotations. The second dictionary maps error names to the number of errors that were
            encountered while decoding the annotations.
        round_precision: The number of digits to round the results to. If None, no rounding is
            performed.
        error_key_correct: The key in the error dictionary whose value should be the number of *correctly*
            decoded annotations, so that the sum of all values in the error dictionary can be used to
            normalize the error counts. If None, the total number of training examples is used to
            normalize the error counts.
        collect_exact_encoding_matches: Whether to collect the number of examples where the full target encoding
            was predicted correctly (exact matches).
    """

    def __init__(
        self,
        layer_metrics: Dict[str, Metric],
        unbatch_function: Callable[[T], Sequence[U]],
        decode_layers_with_errors_function: Callable[[U], Tuple[Dict[str, Any], Dict[str, int]]],
        round_precision: Optional[int] = 4,
        error_key_correct: Optional[str] = None,
        collect_exact_encoding_matches: bool = True,
    ):
        super().__init__()

        self.key_error_correct = error_key_correct
        self.collect_exact_encoding_matches = collect_exact_encoding_matches
        self.round_precision = round_precision
        self.unbatch_function = unbatch_function
        self.decode_layers_with_errors_function = decode_layers_with_errors_function
        self.layer_metrics = ModuleDict(layer_metrics)

        # total number of encodings
        self.add_state("total", default=torch.tensor(0), dist_reduce_fx="sum")
        # this contains the number of examples where the full target sequence was predicted correctly (exact matches)
        self.add_state("exact_encoding_matches", default=torch.tensor(0), dist_reduce_fx="sum")
        # note: the error counts are stored via the MetricWithArbitraryCounts base class

    def update(self, prediction, expected):
        prediction_list = self.unbatch_function(prediction)
        expected_list = self.unbatch_function(expected)
        if len(prediction_list) != len(expected_list):
            raise ValueError(
                f"Number of predictions ({len(prediction_list)}) and targets ({len(expected_list)}) do not match."
            )

        for expected_encoding, prediction_encoding in zip(expected_list, prediction_list):
            expected_layers, _ = self.decode_layers_with_errors_function(expected_encoding)
            predicted_layers, predicted_errors = self.decode_layers_with_errors_function(
                prediction_encoding
            )
            for k, v in predicted_errors.items():
                self.inc_counts(counts=torch.tensor(v).to(self.device), key=k, prefix="errors_")

            for layer_name, metric in self.layer_metrics.items():
                metric.update(expected_layers[layer_name], predicted_layers[layer_name])

            if self.collect_exact_encoding_matches:
                if isinstance(expected_encoding, torch.Tensor) and isinstance(
                    prediction_encoding, torch.Tensor
                ):
                    is_match = torch.equal(expected_encoding, prediction_encoding)
                else:
                    is_match = expected_encoding == prediction_encoding
                if is_match:
                    self.exact_encoding_matches += 1

            self.total += 1

    def reset(self):
        super().reset()

        for metric in self.layer_metrics.values():
            metric.reset()

    def _nested_round(self, d: Dict[str, Any]) -> Dict[str, Any]:
        if self.round_precision is None:
            return d
        res: Dict[str, Any] = {}
        for k, v in d.items():
            if isinstance(v, dict):
                res[k] = self._nested_round(v)
            elif isinstance(v, float):
                res[k] = round(v, self.round_precision)
            else:
                res[k] = v
        return res

    def compute(self):
        res = {}
        if self.collect_exact_encoding_matches:
            res["exact_encoding_matches"] = (
                self.exact_encoding_matches / self.total if self.total > 0 else 0.0
            )

        errors = self.get_counts(key_prefix="errors_")
        # if errors contains a "correct" key, use that to normalize, otherwise use the number of training examples
        if self.key_error_correct in errors:
            errors_total = sum(errors.values())
        else:
            errors_total = self.total
        res["decoding_errors"] = {
            k: v / errors_total if errors_total > 0 else 0.0 for k, v in errors.items()
        }
        if "all" not in res["decoding_errors"]:
            res["decoding_errors"]["all"] = (
                sum(v for k, v in errors.items() if k != self.key_error_correct) / errors_total
                if errors_total > 0
                else 0.0
            )

        for layer_name, metric in self.layer_metrics.items():
            if layer_name in res:
                raise ValueError(
                    f"Layer name '{layer_name}' is already used in the metric result dictionary."
                )
            res[layer_name] = metric.compute()

        res = self._nested_round(res)

        return res



import logging
from abc import ABC
from typing import Dict, Optional

import torch
from torch import LongTensor, Tensor
from torchmetrics import Metric

logger = logging.getLogger(__name__)


class MetricWithArbitraryCounts(Metric, ABC):
    """A metric that hold counts for arbitrary keys."""

    def inc_counts(self, counts: LongTensor, key: Optional[str], prefix: str = "counts_"):
        full_key = prefix
        if key is not None:
            full_key += key

        if not hasattr(self, full_key):
            self.add_state(full_key, default=torch.zeros_like(counts), dist_reduce_fx="sum")

        prev_value = getattr(self, full_key)
        setattr(self, full_key, prev_value + counts)

    def get_counts(self, key_prefix: str = "counts_") -> Dict[Optional[str], LongTensor]:
        result: Dict[Optional[str], LongTensor] = {}
        for k, v in self.metric_state.items():
            if k.startswith(key_prefix):
                if not isinstance(v, Tensor):
                    raise ValueError(
                        f"Expected metric state for key {k} to be a LongTensor, but got {type(v)}."
                    )
                if not isinstance(v, LongTensor):
                    v = v.long()
                assert isinstance(v, LongTensor)
                key = k[len(key_prefix) :] or None
                result[key] = v
        return result



import logging
from typing import Iterator, MutableMapping, Optional, Tuple, Union

import torch.nn
from pie_core import Model
from torch import FloatTensor, LongTensor
from torch.nn import Parameter
from torch.optim import AdamW
from transformers import (
    AutoConfig,
    AutoModelForSequenceClassification,
    get_linear_schedule_with_warmup,
)
from transformers.modeling_outputs import SequenceClassifierOutput
from typing_extensions import TypeAlias

from pytorch_ie.models.common import ModelWithBoilerplate
from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

# model inputs / outputs / targets
InputType: TypeAlias = MutableMapping[str, LongTensor]
OutputType: TypeAlias = SequenceClassifierOutput
TargetType: TypeAlias = MutableMapping[str, Union[LongTensor, FloatTensor]]
# step inputs (batch) / outputs (loss)
StepInputType: TypeAlias = Tuple[InputType, TargetType]
StepOutputType: TypeAlias = FloatTensor


logger = logging.getLogger(__name__)


@Model.register()
class SimpleSequenceClassificationModel(
    ModelWithBoilerplate[InputType, OutputType, TargetType, StepOutputType],
    RequiresModelNameOrPath,
    RequiresNumClasses,
):
    """A simple sequence classification model. It wraps a HuggingFace
    AutoModelForSequenceClassification and adds boilerplate code for training and inference.

    Args:
        model_name_or_path: The name or path of the HuggingFace model to use.
        num_classes: The number of classes for the classification task.
        tokenizer_vocab_size: The size of the tokenizer vocabulary. If provided, the model's
            tokenizer embeddings are resized to this size.
        learning_rate: The learning rate for the optimizer.
        task_learning_rate: The learning rate for the task-specific parameters. If None, the
            learning rate for all parameters is set to `learning_rate`.
        warmup_proportion: The proportion of steps to warm up the learning rate.
        freeze_base_model: If True, the base model parameters are frozen.
        base_model_prefix: The prefix of the base model parameters when using a task_learning_rate
            or freeze_base_model. If None, the base_model_prefix of the model is used.
        **kwargs: Additional keyword arguments passed to the parent class,
            see :class:`ModelWithBoilerplate`.
    """

    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        tokenizer_vocab_size: Optional[int] = None,
        learning_rate: float = 1e-5,
        task_learning_rate: Optional[float] = None,
        warmup_proportion: float = 0.1,
        freeze_base_model: bool = False,
        base_model_prefix: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.save_hyperparameters()

        self.learning_rate = learning_rate
        self.task_learning_rate = task_learning_rate
        self.warmup_proportion = warmup_proportion
        self.freeze_base_model = freeze_base_model

        config = AutoConfig.from_pretrained(model_name_or_path, num_labels=num_classes)
        if self.is_from_pretrained:
            self.model = AutoModelForSequenceClassification.from_config(config=config)
        else:
            self.model = AutoModelForSequenceClassification.from_pretrained(
                model_name_or_path, config=config
            )

        self.base_model_prefix = base_model_prefix or self.model.base_model_prefix

        if tokenizer_vocab_size is not None:
            self.model.resize_token_embeddings(tokenizer_vocab_size)

        if self.freeze_base_model:
            for name, param in self.base_model_named_parameters():
                param.requires_grad = False

    def base_model_named_parameters(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        base_model: Optional[torch.nn.Module] = getattr(self.model, self.base_model_prefix, None)
        if base_model is None:
            raise ValueError(
                f"Base model with prefix '{self.base_model_prefix}' not found in {type(self.model).__name__}"
            )
        if prefix:
            prefix = f"{prefix}."
        return base_model.named_parameters(prefix=f"{prefix}model.{self.base_model_prefix}")

    def task_named_parameters(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        base_model_parameter_names = dict(self.base_model_named_parameters(prefix=prefix)).keys()
        for name, param in self.named_parameters(prefix=prefix):
            if name not in base_model_parameter_names:
                yield name, param

    def forward(self, inputs: InputType, targets: Optional[TargetType] = None) -> OutputType:
        kwargs = {**inputs, **(targets or {})}
        return self.model(**kwargs)

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        labels = torch.argmax(outputs.logits, dim=-1).to(torch.long)
        probabilities = torch.softmax(outputs.logits, dim=-1)
        assert isinstance(labels, LongTensor)
        assert isinstance(probabilities, FloatTensor)
        return {"labels": labels, "probabilities": probabilities}

    def configure_optimizers(self):
        if self.task_learning_rate is not None:
            base_model_params = [param for name, param in self.base_model_named_parameters()]
            task_params = [param for name, param in self.task_named_parameters()]
            optimizer = AdamW(
                [
                    {"params": base_model_params, "lr": self.learning_rate},
                    {"params": task_params, "lr": self.task_learning_rate},
                ]
            )
        else:
            optimizer = AdamW(self.parameters(), lr=self.learning_rate)

        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



from .interfaces import AnnotationEncoderDecoder, DecodingException
from .mixins import BatchableMixin, RelationStatisticsMixin, StatisticsMixin
from .taskmodule_with_document_converter import TaskModuleWithDocumentConverter
from .utils import get_first_occurrence_index



from .has_taskmodule import HasTaskmodule
from .model_with_boilerplate import ModelWithBoilerplate
from .model_with_metrics_from_taskmodule import ModelWithMetricsFromTaskModule
from .stages import TESTING, TRAINING, VALIDATION



from .cross_text_binary_coref import CrossTextBinaryCorefTaskModule
from .labeled_span_extraction_by_token_classification import (
    LabeledSpanExtractionByTokenClassificationTaskModule,
)
from .pointer_network_for_end2end_re import PointerNetworkTaskModuleForEnd2EndRE
from .re_text_classification_with_indices import RETextClassificationWithIndicesTaskModule
from .simple_transformer_text_classification import SimpleTransformerTextClassificationTaskModule
from .text_to_text import TextToTextTaskModule
from .transformer_re_text_classification import TransformerRETextClassificationTaskModule
from .transformer_seq2seq import TransformerSeq2SeqTaskModule
from .transformer_span_classification import TransformerSpanClassificationTaskModule
from .transformer_text_classification import TransformerTextClassificationTaskModule
from .transformer_token_classification import TransformerTokenClassificationTaskModule

__all__ = [
    "SimpleTransformerTextClassificationTaskModule",
    "TransformerRETextClassificationTaskModule",
    "TransformerSeq2SeqTaskModule",
    "TransformerSpanClassificationTaskModule",
    "TransformerTextClassificationTaskModule",
    "TransformerTokenClassificationTaskModule",
    "RETextClassificationWithIndicesTaskModule",
    "TextToTextTaskModule",
    "LabeledSpanExtractionByTokenClassificationTaskModule",
    "PointerNetworkTaskModuleForEnd2EndRE",
    "CrossTextBinaryCorefTaskModule",
]



import logging
from typing import Any, Dict, Generic, Mapping, Tuple, TypeVar

from torch import Tensor
from typing_extensions import TypeAlias

from .model_with_metrics_from_taskmodule import ModelWithMetricsFromTaskModule
from .stages import TESTING, TRAINING, VALIDATION

InputType = TypeVar("InputType")
OutputType = TypeVar("OutputType")
TargetType = TypeVar("TargetType")
StepInputType: TypeAlias = Tuple[InputType, TargetType]
# bound is defined by lightning's training/test_step return type
StepOutputType = TypeVar("StepOutputType", bound=Tensor | Mapping[str, Tensor] | None)

logger = logging.getLogger(__name__)


class ModelWithBoilerplate(
    ModelWithMetricsFromTaskModule[InputType, TargetType, OutputType],
    Generic[InputType, OutputType, TargetType, StepOutputType],
):
    """A PyTorchIEModel that adds boilerplate code for training, validation, and testing.

    Especially, it handles updating the metrics and logging of losses and metric results. Also see
    ModelWithMetricsFromTaskModule for more details on how metrics are handled.
    """

    def get_loss_from_outputs(self, outputs: OutputType) -> StepOutputType:
        if hasattr(outputs, "loss"):
            return outputs.loss
        else:
            raise ValueError(
                f"The model {self.__class__.__name__} does not define a 'loss' attribute in its output, "
                "so the loss cannot be automatically extracted from the outputs. Please override the"
                "get_loss_from_outputs() method for this model."
            )

    def log_loss(self, stage: str, loss: StepOutputType) -> None:
        # show loss on each step only during training
        log_kwargs: Dict[str, Any] = dict(
            on_step=(stage == TRAINING), on_epoch=True, prog_bar=True, sync_dist=True
        )
        if isinstance(loss, Tensor):
            self.log(f"loss/{stage}", loss, **log_kwargs)
        elif isinstance(loss, Mapping):
            for k, v in loss.items():
                if isinstance(v, Tensor):
                    self.log(f"loss/{k}/{stage}", v, **log_kwargs)
                else:
                    logger.warning(
                        f"The loss item '{k}' is of type {type(v)}, but only Tensor losses can be logged."
                    )

    def _step(self, stage: str, batch: StepInputType) -> StepOutputType:
        inputs, targets = batch
        outputs = self(inputs=inputs, targets=targets)
        loss = self.get_loss_from_outputs(outputs=outputs)
        self.log_loss(stage=stage, loss=loss)
        self.update_metric(inputs=inputs, outputs=outputs, targets=targets, stage=stage)

        return loss

    def training_step(self, batch: StepInputType, batch_idx: int) -> StepOutputType:
        return self._step(stage=TRAINING, batch=batch)

    def validation_step(self, batch: StepInputType, batch_idx: int) -> StepOutputType:
        return self._step(stage=VALIDATION, batch=batch)

    def test_step(
        self, batch: StepInputType, batch_idx: int, *args: Any, **kwargs: Any
    ) -> StepOutputType:
        return self._step(stage=TESTING, batch=batch)

    def predict_step(
        self, batch: StepInputType, batch_idx: int, dataloader_idx: int = 0
    ) -> TargetType:
        inputs, targets = batch
        return self.predict(inputs=inputs)

    def on_train_epoch_end(self) -> None:
        self.log_metric(stage=TRAINING)

    def on_validation_epoch_end(self) -> None:
        self.log_metric(stage=VALIDATION)

    def on_test_epoch_end(self) -> None:
        self.log_metric(stage=TESTING)



import copy
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
)

import torch
from pie_core import Annotation, TaskEncoding, TaskModule
from pie_core.utils.dictionary import list_of_dicts2dict_of_lists
from pie_documents.annotations import Span
from pie_documents.documents import TextPairDocumentWithLabeledSpansAndBinaryCorefRelations
from torchmetrics import MetricCollection
from torchmetrics.classification import BinaryAUROC, BinaryAveragePrecision, BinaryF1Score
from transformers import AutoTokenizer, BatchEncoding
from typing_extensions import TypeAlias

from pytorch_ie.taskmodules.common.mixins import RelationStatisticsMixin
from pytorch_ie.taskmodules.metrics import WrappedMetricWithPrepareFunction
from pytorch_ie.utils.document import SpanNotAlignedWithTokenException, get_aligned_token_span
from pytorch_ie.utils.window import get_window_around_slice

logger = logging.getLogger(__name__)

InputEncodingType: TypeAlias = Dict[str, Any]
TargetEncodingType: TypeAlias = Sequence[float]
DocumentType: TypeAlias = TextPairDocumentWithLabeledSpansAndBinaryCorefRelations

TaskEncodingType: TypeAlias = TaskEncoding[
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
]


class TaskOutputType(TypedDict, total=False):
    score: float
    is_similar: bool


ModelInputType: TypeAlias = Dict[str, torch.Tensor]
ModelTargetType: TypeAlias = Dict[str, torch.Tensor]
ModelOutputType: TypeAlias = Dict[str, torch.Tensor]

TaskModuleType: TypeAlias = TaskModule[
    # _InputEncoding, _TargetEncoding, _TaskBatchEncoding, _ModelBatchOutput, _TaskOutput
    DocumentType,
    InputEncodingType,
    TargetEncodingType,
    Tuple[ModelInputType, Optional[ModelTargetType]],
    ModelTargetType,
    TaskOutputType,
]


class SpanDoesNotFitIntoAvailableWindow(Exception):
    def __init__(self, span):
        self.span = span


def _get_labels(model_output: ModelTargetType, label_threshold: float) -> torch.Tensor:
    return (model_output["scores"] > label_threshold).to(torch.int)


def _get_scores(model_output: ModelTargetType) -> torch.Tensor:
    return model_output["scores"]


S = TypeVar("S", bound=Span)


def shift_span(span: S, offset: int) -> S:
    return span.copy(start=span.start + offset, end=span.end + offset)


@TaskModule.register()
class CrossTextBinaryCorefTaskModule(RelationStatisticsMixin, TaskModuleType):
    """This taskmodule processes documents of type
    TextPairDocumentWithLabeledSpansAndBinaryCorefRelations in preparation for a
    SequencePairSimilarityModelWithPooler."""

    DOCUMENT_TYPE = DocumentType

    def __init__(
        self,
        tokenizer_name_or_path: str,
        similarity_threshold: float = 0.9,
        max_window: Optional[int] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_name_or_path)
        self.similarity_threshold = similarity_threshold
        self.max_window = max_window if max_window is not None else self.tokenizer.model_max_length
        self.available_window = self.max_window - self.tokenizer.num_special_tokens_to_add()
        self.num_special_tokens_before = len(self._get_special_tokens_before_input())

    def _get_special_tokens_before_input(self) -> List[int]:
        dummy_ids = self.tokenizer.build_inputs_with_special_tokens(token_ids_0=[-1])
        return dummy_ids[: dummy_ids.index(-1)]

    def encode(self, documents: Union[DocumentType, Iterable[DocumentType]], **kwargs):
        self.reset_statistics()
        result = super().encode(documents=documents, **kwargs)
        self.show_statistics()
        return result

    def truncate_encoding_around_span(
        self, encoding: BatchEncoding, char_span: Span
    ) -> Tuple[Dict[str, List[int]], Span]:
        input_ids = copy.deepcopy(encoding["input_ids"])

        token_span = get_aligned_token_span(encoding=encoding, char_span=char_span)

        # truncate input_ids and shift token_start and token_end
        if len(input_ids) > self.available_window:
            window_slice = get_window_around_slice(
                slice=(token_span.start, token_span.end),
                max_window_size=self.available_window,
                available_input_length=len(input_ids),
            )
            if window_slice is None:
                raise SpanDoesNotFitIntoAvailableWindow(span=token_span)
            window_start, window_end = window_slice
            input_ids = input_ids[window_start:window_end]
            token_span = shift_span(token_span, offset=-window_start)

        truncated_encoding = self.tokenizer.prepare_for_model(ids=input_ids)
        # shift indices because we added special tokens to the input_ids
        token_span = shift_span(token_span, offset=self.num_special_tokens_before)

        return truncated_encoding, token_span

    def encode_input(
        self,
        document: DocumentType,
        is_training: bool = False,
    ) -> Optional[Union[TaskEncodingType, Sequence[TaskEncodingType]]]:
        self.collect_all_relations(kind="available", relations=document.binary_coref_relations)
        tokenizer_kwargs = dict(
            padding=False,
            truncation=False,
            add_special_tokens=False,
        )
        encoding = self.tokenizer(text=document.text, **tokenizer_kwargs)
        encoding_pair = self.tokenizer(text=document.text_pair, **tokenizer_kwargs)

        task_encodings = []
        for coref_rel in document.binary_coref_relations:
            # TODO: This can miss instances if both texts are the same. We could check that
            #   coref_rel.head is in document.labeled_spans (same for the tail), but would this
            #   slow down the encoding?
            if not (
                coref_rel.head.target == document.text
                or coref_rel.tail.target == document.text_pair
            ):
                raise ValueError(
                    f"It is expected that coref relations go from (head) spans over 'text' "
                    f"to (tail) spans over 'text_pair', but this is not the case for this "
                    f"relation (i.e. it points into the other direction): {coref_rel.resolve()}"
                )
            try:
                current_encoding, token_span = self.truncate_encoding_around_span(
                    encoding=encoding, char_span=coref_rel.head
                )
                current_encoding_pair, token_span_pair = self.truncate_encoding_around_span(
                    encoding=encoding_pair, char_span=coref_rel.tail
                )
            except SpanNotAlignedWithTokenException as e:
                logger.warning(
                    f"Could not get token offsets for argument ({e.span}) of coref relation: "
                    f"{coref_rel.resolve()}. Skip it."
                )
                self.collect_relation(kind="skipped_args_not_aligned", relation=coref_rel)
                continue
            except SpanDoesNotFitIntoAvailableWindow as e:
                logger.warning(
                    f"Argument span [{e.span}] does not fit into available token window "
                    f"({self.available_window}). Skip it."
                )
                self.collect_relation(
                    kind="skipped_span_does_not_fit_into_window", relation=coref_rel
                )
                continue

            task_encodings.append(
                TaskEncoding(
                    document=document,
                    inputs={
                        "encoding": current_encoding,
                        "encoding_pair": current_encoding_pair,
                        "pooler_start_indices": token_span.start,
                        "pooler_end_indices": token_span.end,
                        "pooler_pair_start_indices": token_span_pair.start,
                        "pooler_pair_end_indices": token_span_pair.end,
                    },
                    metadata={"candidate_annotation": coref_rel},
                )
            )
            self.collect_relation("used", coref_rel)
        return task_encodings

    def encode_target(
        self,
        task_encoding: TaskEncoding[DocumentType, InputEncodingType, TargetEncodingType],
    ) -> Optional[TargetEncodingType]:
        return task_encoding.metadata["candidate_annotation"].score

    def collate(
        self,
        task_encodings: Sequence[
            TaskEncoding[DocumentType, InputEncodingType, TargetEncodingType]
        ],
    ) -> Tuple[ModelInputType, Optional[ModelTargetType]]:
        inputs_dict = list_of_dicts2dict_of_lists(
            [task_encoding.inputs for task_encoding in task_encodings]
        )

        inputs = {
            k: (
                self.tokenizer.pad(v, return_tensors="pt").data
                if k in ["encoding", "encoding_pair"]
                else torch.tensor(v)
            )
            for k, v in inputs_dict.items()
        }
        for k, v in inputs.items():
            if k.startswith("pooler_") and k.endswith("_indices"):
                inputs[k] = v.unsqueeze(-1)

        if not task_encodings[0].has_targets:
            return inputs, None
        targets = {
            "scores": torch.tensor([task_encoding.targets for task_encoding in task_encodings])
        }
        return inputs, targets

    def configure_model_metric(self, stage: str) -> MetricCollection:
        return MetricCollection(
            metrics={
                "continuous": WrappedMetricWithPrepareFunction(
                    metric=MetricCollection(
                        {
                            "auroc": BinaryAUROC(),
                            "avg-P": BinaryAveragePrecision(validate_args=False),
                            # "roc": BinaryROC(validate_args=False),
                            # "PRCurve": BinaryPrecisionRecallCurve(validate_args=False),
                            "f1": BinaryF1Score(threshold=self.similarity_threshold),
                        }
                    ),
                    prepare_function=_get_scores,
                ),
            }
        )

    def unbatch_output(self, model_output: ModelTargetType) -> Sequence[TaskOutputType]:
        is_similar = (model_output["scores"] > self.similarity_threshold).detach().cpu().tolist()
        scores = model_output["scores"].detach().cpu().tolist()
        result: List[TaskOutputType] = [
            {"is_similar": is_sim, "score": prob} for is_sim, prob in zip(is_similar, scores)
        ]
        return result

    def create_annotations_from_output(
        self,
        task_encoding: TaskEncoding[DocumentType, InputEncodingType, TargetEncodingType],
        task_output: TaskOutputType,
    ) -> Iterator[Tuple[str, Annotation]]:
        if task_output["is_similar"]:
            score = task_output["score"]
            new_coref_rel = task_encoding.metadata["candidate_annotation"].copy(score=score)
            yield "binary_coref_relations", new_coref_rel



from typing import Any, Dict, Optional

from pie_core import AutoTaskModule, TaskModule

from ..interface import RequiresTaskmoduleConfig


class HasTaskmodule(RequiresTaskmoduleConfig):
    """A mixin class for models that have a taskmodule.

    Args:
        taskmodule_config: The config for the taskmodule which can be obtained from the
            taskmodule.config property.
    """

    def __init__(self, taskmodule_config: Optional[Dict[str, Any]] = None, **kwargs):
        super().__init__(**kwargs)
        self.taskmodule: Optional[TaskModule] = None
        if taskmodule_config is not None:
            self.taskmodule = AutoTaskModule.from_config(taskmodule_config)



import abc
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

from pie_core import Annotation

# Annotation Encoding type: encoding for a single annotation
AE = TypeVar("AE")
# Annotation type
A = TypeVar("A", bound=Annotation)
# Annotation Collection Encoding type: encoding for a collection of annotations,
# e.g. all relevant annotations for a document
ACE = TypeVar("ACE")


class DecodingException(Exception, Generic[AE], abc.ABC):
    """Exception raised when decoding fails."""

    identifier: str

    def __init__(self, message: str, encoding: AE):
        self.message = message
        self.encoding = encoding


class AnnotationEncoderDecoder(abc.ABC, Generic[A, AE]):
    """Base class for annotation encoders and decoders."""

    @abc.abstractmethod
    def encode(self, annotation: A, metadata: Optional[Dict[str, Any]] = None) -> AE:
        pass

    @abc.abstractmethod
    def decode(self, encoding: AE, metadata: Optional[Dict[str, Any]] = None) -> A:
        pass

    def build_decoding_constraints(
        self, partial_encoding: AE
    ) -> Tuple[Optional[Any], Optional[Any]]:
        """Given a partial encoding, build the constraints for the next encoding step.

        Returns:
            - A tuple of two elements:
                - The first element is a set of positive constraints for the decoder.
                - The second element is a set of negative constraints for the decoder.
        """
        raise NotImplementedError(
            "build_decoder_constraints is not implemented for this encoder/decoder."
        )

    def parse(self, encoding: AE) -> Tuple[List[A], Dict[str, int], AE]:
        """Parse the encoding and return a list of annotations. This should be error tolerant and
        return all annotations that can be parsed and the remaining encoding.

        Args:
            encoding: The encoding to parse. Can be incomplete.

        Returns:
            - A tuple of three elements:
                - A list of encoded annotations.
                - A dictionary mapping error types to their counts.
                - The remaining encoding after parsing.
        """
        raise NotImplementedError("parse is not implemented for this encoder/decoder.")



import logging
from collections import Counter
from typing import Any, Collection, Dict, Iterable, Optional, Union

import torch
from pie_core import Annotation
from pie_core.utils.dictionary import flatten_dict_s
from torch import FloatTensor, LongTensor

from .common import MetricWithArbitraryCounts

logger = logging.getLogger(__name__)


class PrecisionRecallAndF1ForLabeledAnnotations(MetricWithArbitraryCounts):
    """Computes precision, recall and F1 for labeled annotations. Inputs and targets are lists of
    annotations. True positives are counted as the number of annotations that are the same in both
    inputs and targets calculated as exact matches via set operation, false positives and false
    negatives accordingly. The annotations are deduplicated for each instance. But if the same
    annotation occurs in different instances, it is counted as two separate annotations.

    Args:
        label_mapping: A dictionary mapping annotation labels to human-readable labels. If None,
            the annotation labels are used as they are. Can be used to map label ids to string labels.
        key_micro: The key to use for the micro-average in the metric result dictionary.
        in_percent: Whether to return the results in percent, i.e. values between 0 and 100 instead of
            between 0 and 1.
        flatten_result_with_sep: If not None, the result dictionary is flattened and the keys of the
            different nesting levels are concatenated with the given separator.
        prefix: If not None, the most outer keys of the result dictionary are prefixed with this string.
        return_recall_and_precision: Whether to return recall and precision in addition to F1.
    """

    def __init__(
        self,
        label_mapping: Optional[Dict[Any, str]] = None,
        key_micro: Optional[str] = "micro",
        key_macro: Optional[str] = "macro",
        in_percent: bool = False,
        flatten_result_with_sep: Optional[str] = None,
        prefix: Optional[str] = None,
        return_recall_and_precision: bool = True,
    ):
        super().__init__()
        self.label_mapping = label_mapping
        self.key_micro = key_micro
        self.key_macro = key_macro
        self.in_percent = in_percent
        self.flatten_result_with_sep = flatten_result_with_sep
        self.prefix = prefix
        self.return_recall_and_precision = return_recall_and_precision

    def update(self, gold: Iterable[Annotation], predicted: Iterable[Annotation]) -> None:
        # remove duplicates within each list
        gold_set = set(gold)
        predicted_set = set(predicted)
        new_counts = self.calculate_counts(gold_set, predicted_set, gold_set & predicted_set)
        for k, v in new_counts.items():
            self.inc_counts(counts=v, key=k)

    def get_precision_recall_f1(
        self, n_gold_predicted_correct: LongTensor
    ) -> Dict[str, FloatTensor]:
        n_gold = n_gold_predicted_correct[0]
        n_predicted = n_gold_predicted_correct[1]
        n_correct = n_gold_predicted_correct[2]
        zero = torch.tensor(0.0).to(self.device)
        recall = zero if n_gold == 0 else (n_correct / n_gold)
        precision = zero if n_predicted == 0 else (n_correct / n_predicted)
        f1 = zero if recall + precision == 0 else (2 * precision * recall) / (precision + recall)

        result = {"f1": f1}
        if self.return_recall_and_precision:
            result["recall"] = recall
            result["precision"] = precision

        if self.in_percent:
            result = {k: v * 100 for k, v in result.items()}

        assert all(isinstance(v, FloatTensor) for v in result.values())
        # we can not ensure the type of the whole dict, but value types are checked above
        return result  # type: ignore[return-value]

    def get_label(self, annotation: Annotation) -> Optional[str]:
        label: Optional[str] = getattr(annotation, "label", None)
        if self.label_mapping is not None:
            return self.label_mapping[label]
        return label

    def calculate_counts(
        self,
        gold: Collection[Annotation],
        predicted: Collection[Annotation],
        correct: Collection[Annotation],
    ) -> Dict[Optional[str], LongTensor]:
        result = {}
        # per class
        gold_counter = Counter([self.get_label(ann) for ann in gold])
        predicted_counter = Counter([self.get_label(ann) for ann in predicted])
        correct_counter = Counter([self.get_label(ann) for ann in correct])
        for label in gold_counter.keys() | predicted_counter.keys():
            if self.key_micro is not None and label == self.key_micro:
                raise ValueError(
                    f"The key '{self.key_micro}' was used as an annotation label, but it is reserved for "
                    f"the micro average. You can change which key is used for that with the 'key_micro' argument."
                )
            result[label] = torch.tensor(
                [
                    gold_counter.get(label, 0),
                    predicted_counter.get(label, 0),
                    correct_counter.get(label, 0),
                ]
            ).to(device=self.device)

        # overall
        if self.key_micro is not None:
            result[self.key_micro] = torch.tensor([len(gold), len(predicted), len(correct)]).to(
                device=self.device
            )

        assert all(isinstance(v, LongTensor) for v in result.values())
        # we can not ensure the type of the whole dict, but value types are checked above
        return result  # type: ignore[return-value]

    def compute(self) -> dict[str | None, dict[str, FloatTensor]]:
        counts = self.get_counts()
        result = {label: self.get_precision_recall_f1(counts[label]) for label in counts.keys()}
        if self.key_macro is not None:
            result_without_micro = {
                k: v for k, v in result.items() if self.key_micro is None or k != self.key_micro
            }
            if len(result_without_micro) > 0:
                sub_keys = list(result_without_micro.values())[0].keys()
                macro_scores = {
                    k: torch.stack([v[k] for v in result_without_micro.values()]).mean()
                    for k in sub_keys
                }
                assert all(isinstance(v, FloatTensor) for v in macro_scores.values())
                # we can not ensure the type of the whole dict, but value types are checked above
                result[self.key_macro] = macro_scores  # type: ignore[assignment]

        if self.prefix is not None:
            result = {f"{self.prefix}{k}": v for k, v in result.items()}

        if self.flatten_result_with_sep is not None:
            return flatten_dict_s(result, sep=self.flatten_result_with_sep)
        else:
            return result



from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import torch.utils.checkpoint
from torch import nn
from torch.nn import Parameter
from torch.optim import Optimizer
from transformers import BartConfig, BartModel, BartPreTrainedModel, GenerationConfig
from transformers.modeling_outputs import Seq2SeqLMOutput, Seq2SeqModelOutput
from transformers.models.bart.modeling_bart import shift_tokens_right
from transformers.utils import logging

from ..components.pointer_head import PointerHead
from .bart_with_decoder_position_ids import BartModelWithDecoderPositionIds

logger = logging.get_logger(__name__)


def get_layer_norm_parameters(
    named_parameters: Iterator[Tuple[str, Parameter]],
) -> Iterator[Parameter]:
    return (
        param for name, param in named_parameters if "layernorm" in name or "layer_norm" in name
    )


def get_non_layer_norm_parameters(
    named_parameters: Iterator[Tuple[str, Parameter]],
) -> Iterator[Parameter]:
    return (
        param
        for name, param in named_parameters
        if not ("layernorm" in name or "layer_norm" in name)
    )


class BartAsPointerNetworkConfig(BartConfig):
    def __init__(
        self,
        # respective token ids for the label-, eos-, and pad ids. Can be used as a mapping from the
        # target ids to the token ids.
        target_token_ids: Optional[List[int]] = None,
        # token id mapping to better initialize the label embedding weights
        embedding_weight_mapping: Optional[Dict[Union[int, str], List[int]]] = None,
        # special decoder position id handling
        decoder_position_id_mode: Optional[str] = None,
        decoder_position_id_pattern: Optional[List[int]] = None,
        decoder_position_id_mapping: Optional[Dict[int, int]] = None,
        # other parameters
        use_encoder_mlp: bool = True,
        use_constraints_encoder_mlp: bool = False,
        # optimizer
        lr: float = 5e-5,
        task_lr: Optional[float] = None,
        weight_decay: float = 1e-2,
        head_decay: Optional[float] = None,
        shared_decay: Optional[float] = None,
        encoder_layer_norm_decay: Optional[float] = 0.001,
        decoder_layer_norm_decay: Optional[float] = None,
        # other BartConfig parameters
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.target_token_ids = target_token_ids

        self.embedding_weight_mapping = embedding_weight_mapping

        self.use_encoder_mlp = use_encoder_mlp
        self.use_constraints_encoder_mlp = use_constraints_encoder_mlp

        self.decoder_position_id_mode = decoder_position_id_mode
        self.decoder_position_id_pattern = decoder_position_id_pattern
        self.decoder_position_id_mapping = decoder_position_id_mapping

        self.lr = lr
        self.task_lr = task_lr
        self.weight_decay = weight_decay
        self.head_decay = head_decay
        self.shared_decay = shared_decay
        self.encoder_layer_norm_decay = encoder_layer_norm_decay
        self.decoder_layer_norm_decay = decoder_layer_norm_decay


class BartAsPointerNetwork(BartPreTrainedModel):
    config_class = BartAsPointerNetworkConfig
    base_model_prefix = "model"
    _tied_weights_keys = [
        "encoder.embed_tokens.weight",
        "decoder.embed_tokens.weight",
    ]

    def __init__(self, config: BartAsPointerNetworkConfig):
        super().__init__(config)
        if self.config.decoder_position_id_mode is not None:
            self.model = BartModelWithDecoderPositionIds(config)
        else:
            self.model = BartModel(config)

        self.pointer_head = PointerHead(
            # target space ids
            bos_id=self.model.config.bos_token_id,
            eos_id=self.model.config.eos_token_id,
            pad_id=self.model.config.pad_token_id,
            # decoder-input token ids
            target_token_ids=self.model.config.target_token_ids,
            # embeddings
            embeddings=self.model.decoder.embed_tokens,
            embedding_weight_mapping=self.model.config.embedding_weight_mapping,
            # other parameters
            use_encoder_mlp=self.model.config.use_encoder_mlp,
            use_constraints_encoder_mlp=self.model.config.use_constraints_encoder_mlp,
            decoder_position_id_mode=self.model.config.decoder_position_id_mode,
            decoder_position_id_pattern=self.model.config.decoder_position_id_pattern,
            decoder_position_id_mapping=self.model.config.decoder_position_id_mapping,
        )

        # Initialize weights and apply final processing
        self.post_init()

    @classmethod
    def _load_pretrained_model(
        cls,
        *args,
        **kwargs,
    ):
        (
            model,
            missing_keys,
            unexpected_keys,
            mismatched_keys,
            offload_index,
            error_msgs,
        ) = super()._load_pretrained_model(*args, **kwargs)
        # adjust the model after loading the original model (e.g. vanilla BartModel)
        model.adjust_after_loading_original_model()
        return model, missing_keys, unexpected_keys, mismatched_keys, offload_index, error_msgs

    def resize_token_embeddings(
        self, new_num_tokens: Optional[int] = None, pad_to_multiple_of: Optional[int] = None
    ) -> nn.Embedding:
        new_embeddings = super().resize_token_embeddings(new_num_tokens, pad_to_multiple_of)
        # we also need to update the embeddings in the pointer head
        self.pointer_head.set_embeddings(new_embeddings)
        return new_embeddings

    def adjust_after_loading_original_model(self):
        # target_token_ids contains all new target tokens for the labels and new tokens were added to the end
        # of the vocabulary, so we can use its maximum to resize the embedding weights
        self.resize_token_embeddings(new_num_tokens=max(self.config.target_token_ids) + 1)
        # initialize the newly added embeddings for the labels with better weights from the original embeddings
        self.pointer_head.overwrite_embeddings_with_mapping()

        # adjust generation settings
        # set the correct decoder_start_token_id
        self.config.decoder_start_token_id = self.config.bos_token_id
        # disable ForcedBOSTokenLogitsProcessor
        self.config.forced_bos_token_id = None
        # disable ForcedEOSTokenLogitsProcessor
        self.config.forced_eos_token_id = None
        # update the generation config accordingly
        self.generation_config = GenerationConfig.from_model_config(self.config)

    def base_model_named_params(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        yield from self.model.named_parameters(prefix=prefix + self.base_model_prefix)

    def head_named_params(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        base_model_param_names = {
            name for name, param in self.base_model_named_params(prefix=prefix)
        }
        for name, param in self.named_parameters(prefix=prefix):
            if name not in base_model_param_names:
                yield name, param

    def encoder_only_named_params(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        shared_params = set(dict(self.encoder_decoder_shared_named_params(prefix=prefix)).values())
        for name, param in self.model.encoder.named_parameters(
            prefix=prefix + self.base_model_prefix + ".encoder"
        ):
            if param not in shared_params:
                yield name, param

    def decoder_only_named_params(self, prefix: str = "") -> Iterator[Tuple[str, Parameter]]:
        shared_params = set(dict(self.encoder_decoder_shared_named_params(prefix=prefix)).values())
        for name, param in self.model.decoder.named_parameters(
            prefix=prefix + self.base_model_prefix + ".decoder"
        ):
            if param not in shared_params:
                yield name, param

    def encoder_decoder_shared_named_params(
        self, prefix: str = ""
    ) -> Iterator[Tuple[str, Parameter]]:
        encoder_params = set(self.model.encoder.parameters())
        decoder_params = set(self.model.decoder.parameters())
        for name, param in self.base_model_named_params(prefix=prefix):
            if param in encoder_params and param in decoder_params:
                yield name, param

    def get_encoder(self):
        return self.model.get_encoder()

    def get_decoder(self):
        return self.model.get_decoder()

    # @add_start_docstrings_to_model_forward(BART_INPUTS_DOCSTRING)
    # @replace_return_docstrings(output_type=Seq2SeqLMOutput, config_class=_CONFIG_FOR_DOC)
    # @add_end_docstrings(BART_GENERATION_EXAMPLE)
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        decoder_position_ids: Optional[torch.LongTensor] = None,
        constraints: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[List[torch.FloatTensor]] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, Seq2SeqLMOutput]:
        r"""Labels (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*): Labels
        for computing the masked language modeling loss. Indices should either be in `[0, ...,
        config.vocab_size]` or -100 (see `input_ids` docstring). Tokens with indices set to `-100`
        are ignored (masked), the loss is only computed for the tokens with labels in `[0, ...,
        config.vocab_size]`.

        Returns:
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if labels is not None:
            if use_cache:
                logger.warning(
                    "The `use_cache` argument is changed to `False` since `labels` is provided."
                )
            use_cache = False
            if decoder_input_ids is None and decoder_inputs_embeds is None:
                decoder_input_ids = shift_tokens_right(
                    labels, self.config.pad_token_id, self.config.decoder_start_token_id
                )

        if decoder_input_ids is None:
            # we can not create the decoder_input_ids from input_ids, because we need the
            # encoder_input_ids for the pointer network
            raise ValueError("decoder_input_ids has to be set!")

        if input_ids is None:
            raise ValueError("input_ids has to be set!")

        # this adjusts the input_ids and, if available, the position_ids
        decoder_inputs = self.pointer_head.prepare_decoder_inputs(
            input_ids=decoder_input_ids,
            # in the case of generation (with past_key_values) the position_ids are already prepared
            position_ids=decoder_position_ids,
            encoder_input_ids=input_ids,
        )

        model_inputs = dict(
            input_ids=input_ids,
            encoder_outputs=encoder_outputs,
            attention_mask=attention_mask,
            decoder_attention_mask=decoder_attention_mask,
            head_mask=head_mask,
            decoder_head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            decoder_inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=True,
        )
        for k, v in decoder_inputs.items():
            model_inputs[f"decoder_{k}"] = v
        outputs = self.model(**model_inputs)

        if not isinstance(outputs, Seq2SeqModelOutput):
            raise ValueError(
                "Inconsistent output: The output of the model forward should be of type "
                f"`Seq2SeqLMOutput`, but is of type `{type(outputs)}`."
            )
        logits, loss = self.pointer_head(
            last_hidden_state=outputs.last_hidden_state,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_input_ids=input_ids,
            encoder_attention_mask=attention_mask,
            labels=labels,
            decoder_attention_mask=decoder_attention_mask,
            constraints=constraints,
        )

        if not return_dict:
            output = (logits,) + outputs[1:]
            return ((loss,) + output) if loss is not None else output

        return Seq2SeqLMOutput(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            decoder_hidden_states=outputs.decoder_hidden_states,
            decoder_attentions=outputs.decoder_attentions,
            cross_attentions=outputs.cross_attentions,
            encoder_last_hidden_state=outputs.encoder_last_hidden_state,
            encoder_hidden_states=outputs.encoder_hidden_states,
            encoder_attentions=outputs.encoder_attentions,
        )

    def prepare_inputs_for_generation(
        self,
        decoder_input_ids,
        encoder_input_ids,  # added for pointer network
        encoder_attention_mask,  # added for pointer network
        past_key_values=None,
        attention_mask=None,
        decoder_attention_mask=None,
        head_mask=None,
        decoder_head_mask=None,
        cross_attn_head_mask=None,
        use_cache=None,
        encoder_outputs=None,
        **kwargs,
    ):
        result = {}
        if self.pointer_head.use_prepared_position_ids:
            # we need to prepare the position ids for the decoder here, because later we do not have the full
            # input_ids anymore
            result["decoder_position_ids"] = self.pointer_head.prepare_decoder_position_ids(
                input_ids=decoder_input_ids
            )

        # cut decoder_input_ids if past_key_values is used
        if past_key_values is not None:
            past_length = past_key_values[0][0].shape[2]

            # Some generation methods already pass only the last input ID
            if decoder_input_ids.shape[1] > past_length:
                remove_prefix_length = past_length
            else:
                # Default to old behavior: keep only final ID
                remove_prefix_length = decoder_input_ids.shape[1] - 1

            decoder_input_ids = decoder_input_ids[:, remove_prefix_length:]

            if "decoder_position_ids" in result:
                result["decoder_position_ids"] = result["decoder_position_ids"][
                    :, remove_prefix_length:
                ]

        result.update(
            {
                "input_ids": encoder_input_ids,
                "encoder_outputs": encoder_outputs,
                "past_key_values": past_key_values,
                "decoder_input_ids": decoder_input_ids,
                "attention_mask": encoder_attention_mask,
                "decoder_attention_mask": decoder_attention_mask,
                "head_mask": head_mask,
                "decoder_head_mask": decoder_head_mask,
                "cross_attn_head_mask": cross_attn_head_mask,
                "use_cache": use_cache,  # change this to avoid caching (presumably for debugging)
            }
        )
        return result

    def prepare_decoder_input_ids_from_labels(self, labels: torch.Tensor):
        return shift_tokens_right(
            labels, self.config.pad_token_id, self.config.decoder_start_token_id
        )

    @staticmethod
    def _reorder_cache(past_key_values, beam_idx):
        reordered_past = ()
        for layer_past in past_key_values:
            # cached cross_attention states don't have to be reordered -> they are always the same
            reordered_past += (
                tuple(
                    past_state.index_select(0, beam_idx.to(past_state.device))
                    for past_state in layer_past[:2]
                )
                + layer_past[2:],
            )
        return reordered_past

    def _prepare_encoder_decoder_kwargs_for_generation(
        self, inputs_tensor: torch.Tensor, model_kwargs, model_input_name: Optional[str] = None
    ) -> Dict[str, Any]:
        result = super()._prepare_encoder_decoder_kwargs_for_generation(
            inputs_tensor=inputs_tensor,
            model_kwargs=model_kwargs,
            model_input_name=model_input_name,
        )
        # add items that are needed for pointer network
        result["encoder_input_ids"] = inputs_tensor
        result["encoder_attention_mask"] = result["attention_mask"]
        return result

    def configure_optimizer(self) -> Optimizer:
        parameters = []

        # head parameters
        head_decay = (
            self.config.head_decay
            if self.config.head_decay is not None
            else self.config.weight_decay
        )
        params = {
            "lr": self.config.task_lr if self.config.task_lr is not None else self.config.lr,
            "weight_decay": head_decay,
            "params": dict(self.head_named_params()).values(),
        }
        parameters.append(params)

        # decoder only layer norm parameters
        decoder_layer_norm_decay = (
            self.config.decoder_layer_norm_decay
            if self.config.decoder_layer_norm_decay is not None
            else self.config.weight_decay
        )
        params = {
            "lr": self.config.lr,
            "weight_decay": decoder_layer_norm_decay,
            "params": get_layer_norm_parameters(self.decoder_only_named_params()),
        }
        parameters.append(params)

        # decoder only other parameters
        params = {
            "lr": self.config.lr,
            "weight_decay": self.config.weight_decay,
            "params": get_non_layer_norm_parameters(self.decoder_only_named_params()),
        }
        parameters.append(params)

        # encoder only layer norm parameters
        encoder_layer_norm_decay = (
            self.config.encoder_layer_norm_decay
            if self.config.encoder_layer_norm_decay is not None
            else self.config.weight_decay
        )
        params = {
            "lr": self.config.lr,
            "weight_decay": encoder_layer_norm_decay,
            "params": get_layer_norm_parameters(self.encoder_only_named_params()),
        }
        parameters.append(params)

        # encoder only other parameters
        params = {
            "lr": self.config.lr,
            "weight_decay": self.config.weight_decay,
            "params": get_non_layer_norm_parameters(self.encoder_only_named_params()),
        }
        parameters.append(params)

        # encoder-decoder shared parameters
        shared_decay = (
            self.config.shared_decay
            if self.config.shared_decay is not None
            else self.config.weight_decay
        )
        params = {
            "lr": self.config.lr,
            "weight_decay": shared_decay,
            "params": dict(self.encoder_decoder_shared_named_params()).values(),
        }
        parameters.append(params)

        return torch.optim.AdamW(parameters)



from pytorch_ie.models.sequence_classification_with_pooler import (
    SequenceClassificationModelWithPooler,
    SequencePairSimilarityModelWithPooler,
)
from pytorch_ie.models.simple_generative import SimpleGenerativeModel
from pytorch_ie.models.simple_sequence_classification import SimpleSequenceClassificationModel
from pytorch_ie.models.simple_token_classification import SimpleTokenClassificationModel
from pytorch_ie.models.token_classification_with_seq2seq_encoder_and_crf import (
    TokenClassificationModelWithSeq2SeqEncoderAndCrf,
)
from pytorch_ie.models.transformer_seq2seq import TransformerSeq2SeqModel
from pytorch_ie.models.transformer_span_classification import TransformerSpanClassificationModel
from pytorch_ie.models.transformer_text_classification import TransformerTextClassificationModel
from pytorch_ie.models.transformer_token_classification import TransformerTokenClassificationModel

__all__ = [
    "TransformerSeq2SeqModel",
    "TransformerSpanClassificationModel",
    "TransformerTextClassificationModel",
    "TransformerTokenClassificationModel",
    "SequenceClassificationModelWithPooler",
    "SequencePairSimilarityModelWithPooler",
    "SimpleTokenClassificationModel",
    "SimpleGenerativeModel",
    "SimpleSequenceClassificationModel",
    "TokenClassificationModelWithSeq2SeqEncoderAndCrf",
]



import logging
from typing import MutableMapping, Optional, Tuple, Union

import torch
from pie_core import Model
from pytorch_lightning.utilities.types import OptimizerLRScheduler
from torch import FloatTensor, LongTensor
from transformers import AutoConfig, AutoModelForTokenClassification, BatchEncoding
from transformers.modeling_outputs import TokenClassifierOutput
from typing_extensions import TypeAlias

from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

from .common import ModelWithBoilerplate

# model inputs / outputs / targets
InputType: TypeAlias = BatchEncoding
OutputType: TypeAlias = TokenClassifierOutput
TargetType: TypeAlias = MutableMapping[str, Union[LongTensor, FloatTensor]]
# step inputs (batch) / outputs (loss)
StepInputType: TypeAlias = Tuple[InputType, TargetType]
StepOutputType: TypeAlias = FloatTensor


logger = logging.getLogger(__name__)


@Model.register()
class SimpleTokenClassificationModel(
    ModelWithBoilerplate[InputType, OutputType, TargetType, StepOutputType],
    RequiresModelNameOrPath,
    RequiresNumClasses,
):
    """A simple token classification model that wraps a (pretrained) model loaded with
    AutoModelForTokenClassification from the transformers library.

    The model is trained with a cross-entropy loss function and uses the Adam optimizer.

    Note that for training, the labels for the special tokens (as well as for padding tokens)
    are expected to have the value label_pad_id (-100 by default, which is the default ignore_index
    value for the CrossEntropyLoss). The predictions for these tokens are also replaced with
    label_pad_id to match the training labels for correct metric calculation. Therefore, the model
    requires the special_tokens_mask and attention_mask (for padding) to be passed as inputs.

    Args:
        model_name_or_path: The name or path of the pretrained transformer model to use.
        num_classes: The number of classes to predict.
        learning_rate: The learning rate to use for training.
        label_pad_id: The label id to use for padding labels (at the padding token positions
            as well as for the special tokens).
    """

    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        learning_rate: float = 1e-5,
        label_pad_id: int = -100,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.learning_rate = learning_rate
        self.label_pad_id = label_pad_id
        self.num_classes = num_classes

        config = AutoConfig.from_pretrained(model_name_or_path, num_labels=num_classes)
        if self.is_from_pretrained:
            self.model = AutoModelForTokenClassification.from_config(config=config)
        else:
            self.model = AutoModelForTokenClassification.from_pretrained(
                model_name_or_path, config=config
            )

    def forward(self, inputs: InputType, targets: Optional[TargetType] = None) -> OutputType:
        inputs_without_special_tokens_mask = {
            k: v for k, v in inputs.items() if k != "special_tokens_mask"
        }
        return self.model(**inputs_without_special_tokens_mask, **(targets or {}))

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        # get the max index for each token from the logits
        tags_tensor = torch.argmax(outputs.logits, dim=-1)

        # mask out the padding and special tokens
        tags_tensor = tags_tensor.masked_fill(inputs["attention_mask"] == 0, self.label_pad_id)

        # mask out the special tokens
        tags_tensor = tags_tensor.masked_fill(
            inputs["special_tokens_mask"] == 1, self.label_pad_id
        )
        labels = tags_tensor.to(torch.long)
        probabilities = torch.softmax(outputs.logits, dim=-1)
        assert isinstance(labels, LongTensor)
        assert isinstance(probabilities, FloatTensor)
        return {"labels": labels, "probabilities": probabilities}

    def configure_optimizers(self) -> OptimizerLRScheduler:
        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)



class RequiresModelNameOrPath:
    pass


class RequiresNumClasses:
    pass


class RequiresMaxInputLength:
    """Any class inheriting from this class should require a constructor parameter
    'max_input_length'."""

    pass


class RequiresTaskmoduleConfig:
    """Any class inheriting from this class should require a constructor parameter
    'taskmodule_config'."""

    pass



from pie_documents.metrics.statistics import (
    DummyCollector,
    FieldLengthCollector,
    LabelCountCollector,
    SubFieldLengthCollector,
)

from .token_count_collector import TokenCountCollector



import logging
from typing import Any, Dict, Optional, Type, Union

from pie_core import Document, DocumentStatistic
from pie_documents.documents import TextBasedDocument
from transformers import AutoTokenizer, PreTrainedTokenizer

logger = logging.getLogger(__name__)


class TokenCountCollector(DocumentStatistic):
    """Collects the token count of a field when tokenizing its content with a Huggingface
    tokenizer.

    The content of the field should be a string.
    """

    def __init__(
        self,
        tokenizer: Union[str, PreTrainedTokenizer],
        text_field: str = "text",
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        document_type: Optional[Type[Document]] = None,
        **kwargs,
    ):
        if document_type is None and text_field == "text":
            document_type = TextBasedDocument
        super().__init__(document_type=document_type, **kwargs)
        self.tokenizer = (
            AutoTokenizer.from_pretrained(tokenizer) if isinstance(tokenizer, str) else tokenizer
        )
        self.tokenizer_kwargs = tokenizer_kwargs or {}
        self.text_field = text_field

    def _collect(self, doc: Document) -> int:
        text = getattr(doc, self.text_field)
        encodings = self.tokenizer(text, **self.tokenizer_kwargs)
        tokens = encodings.tokens()
        return len(tokens)



import logging
from copy import copy
from typing import Any, Dict, List, Optional, Tuple

from torch import Tensor, nn

logger = logging.getLogger(__name__)

RNN_TYPE2CLASS = {"lstm": nn.LSTM, "gru": nn.GRU, "rnn": nn.RNN}
ACTIVATION_TYPE2CLASS = {
    "relu": nn.ReLU,
    "tanh": nn.Tanh,
    "sigmoid": nn.Sigmoid,
    "gelu": nn.GELU,
}


class RNNWrapper(nn.Module):
    def __init__(self, rnn: nn.Module):
        super().__init__()
        self.rnn = rnn

    def forward(self, *args, **kwargs) -> Tensor:
        return self.rnn(*args, **kwargs)[0]

    @property
    def output_size(self) -> int:
        if self.rnn.bidirectional:
            return self.rnn.hidden_size * 2
        else:
            return self.rnn.hidden_size


def build_seq2seq_encoder(
    config: Dict[str, Any], input_size: int
) -> Tuple[Optional[nn.Module], int]:
    # copy the config to avoid side effects
    config = copy(config)
    seq2seq_encoder_type = config.pop("type", None)
    seq2seq_encoder: Optional[nn.Module]
    if seq2seq_encoder_type is None:
        logger.warning(
            f"seq2seq_encoder_type is not specified in the seq2seq_encoder: {config}. "
            f"Do not build this seq2seq_encoder."
        )
        return None, input_size

    if seq2seq_encoder_type == "sequential":
        modules: List[nn.Module] = []
        output_size = input_size
        for key, subconfig in config.items():
            module, output_size = build_seq2seq_encoder(subconfig, input_size)
            if module is not None:
                modules.append(module)
            input_size = output_size

        seq2seq_encoder = nn.Sequential(*modules)
    elif seq2seq_encoder_type in RNN_TYPE2CLASS:
        rnn_class = RNN_TYPE2CLASS[seq2seq_encoder_type]
        seq2seq_encoder = RNNWrapper(rnn_class(input_size=input_size, batch_first=True, **config))
        output_size = seq2seq_encoder.output_size
    elif seq2seq_encoder_type == "linear":
        seq2seq_encoder = nn.Linear(in_features=input_size, **config)
        output_size = seq2seq_encoder.out_features
    elif seq2seq_encoder_type in ACTIVATION_TYPE2CLASS:
        activation_class = ACTIVATION_TYPE2CLASS[seq2seq_encoder_type]
        seq2seq_encoder = activation_class(**config)
        output_size = input_size
    elif seq2seq_encoder_type == "dropout":
        seq2seq_encoder = nn.Dropout(**config)
        output_size = input_size
    elif seq2seq_encoder_type == "none":
        seq2seq_encoder = None
        output_size = input_size
    else:
        raise ValueError(f"Unknown seq2seq_encoder_type: {seq2seq_encoder_type}")

    return seq2seq_encoder, output_size



import logging
from typing import Union

import torch

logger = logging.getLogger(__name__)


def get_first_occurrence_index(
    tensor: Union[torch.FloatTensor, torch.LongTensor], value: Union[float, int]
) -> torch.LongTensor:
    """Returns the index of the first occurrence of `value` in each row of `tensor`. If `value` is
    not found, seq_len is returned.

    Args:
        tensor: the tensor of shape (bsz, seq_len) to search in
        value: the value to search for

    Returns: a tensor of shape (bsz,) containing the index of the first occurrence of `value` in each row of `tensor`.
    """

    mask_value = tensor.eq(value)
    # count matching positions from the end
    value_counts_to_end = mask_value.flip(dims=[1]).cumsum(dim=1).flip(dims=[1])
    # at the first position stands the number of total matches
    total_matches = value_counts_to_end[:, 0]
    # the sum of all positions where the number of matches is equal to the total number of matches
    # is the index *after* the first occurrence
    result = value_counts_to_end.eq(total_matches.unsqueeze(-1)).sum(dim=1) - 1
    # set result to seq_len if no match was found
    result[total_matches == 0] = tensor.size(1)

    assert isinstance(result, torch.LongTensor)
    return result



import logging
from typing import Any, Dict, MutableMapping, Optional, Tuple, Union

import torch
from pie_core import Model
from pytorch_lightning.utilities.types import OptimizerLRScheduler
from torch import FloatTensor, LongTensor, nn
from transformers import AutoConfig, AutoModel, BatchEncoding, get_linear_schedule_with_warmup
from transformers.modeling_outputs import TokenClassifierOutput
from typing_extensions import TypeAlias

from pytorch_ie.models.interface import RequiresModelNameOrPath, RequiresNumClasses

from .common import ModelWithBoilerplate
from .components.seq2seq_encoder import build_seq2seq_encoder

# model inputs / outputs / targets
InputType: TypeAlias = BatchEncoding
OutputType: TypeAlias = TokenClassifierOutput
TargetType: TypeAlias = MutableMapping[str, Union[LongTensor, FloatTensor]]
# step inputs (batch) / outputs (loss)
StepInputType: TypeAlias = Tuple[InputType, TargetType]
StepOutputType: TypeAlias = FloatTensor

HF_MODEL_TYPE_TO_CLASSIFIER_DROPOUT_ATTRIBUTE = {
    "bert": "hidden_dropout_prob",
    "roberta": "hidden_dropout_prob",
    "albert": "classifier_dropout_prob",
    "distilbert": "seq_classif_dropout",
    "deberta-v2": "hidden_dropout_prob",
    "longformer": "hidden_dropout_prob",
}

logger = logging.getLogger(__name__)


@Model.register()
class TokenClassificationModelWithSeq2SeqEncoderAndCrf(
    ModelWithBoilerplate[InputType, OutputType, TargetType, StepOutputType],
    RequiresNumClasses,
    RequiresModelNameOrPath,
):
    """A token classification model that wraps a (pretrained) model loaded with AutoModel from the
    transformers library. The model can optionally be followed by a seq2seq encoder (e.g. an LSTM).
    Finally, Conditional Random Fields (CRFs) can be used to decode the predictions.

    The model is trained with a cross-entropy loss function and uses the Adam optimizer.

    Note that for training, the labels for the special tokens (as well as for padding tokens)
    are expected to have the value label_pad_id (-100 by default, which is the default ignore_index
    value for the CrossEntropyLoss). The predictions for these tokens are also replaced with
    label_pad_id to match the training labels for correct metric calculation. Therefore, the model
    requires the special_tokens_mask and attention_mask (for padding) to be passed as inputs.

    Args:
        model_name_or_path: The name or path of the (pretrained) transformer model to use.
        num_classes: The number of classes to predict.
        learning_rate: The learning rate to use for training.
        task_learning_rate: The learning rate to use for the task-specific parameters, i.e.
            for the sequence-to-sequence encoder, classification head, and CRF. If None, the
            learning_rate is used for all parameters.
        use_crf: Whether to use a CRF to decode the predictions.
        label_pad_id: The label id to use for padding labels (at the padding token positions
            as well as for the special tokens).
        special_token_label_id: The label id to use for special tokens (e.g. [CLS], [SEP]). This
            is used to replace the targets for special tokens with the label_pad_id before passing
            them to the CRF because the CRF does not allow the first token to be masked out.
        classifier_dropout: The dropout probability to use for the classification head.
        freeze_base_model: Whether to freeze the base model (i.e. the transformer) during training.
        warmup_proportion: The proportion of training steps to use for the linear warmup.
        seq2seq_encoder: A dictionary with the configuration for the seq2seq encoder. If None, no
            seq2seq encoder is used. See ./components/seq2seq_encoder.py for further information.
    """

    def __init__(
        self,
        model_name_or_path: str,
        num_classes: int,
        learning_rate: float = 1e-5,
        task_learning_rate: Optional[float] = None,
        use_crf: bool = True,
        label_pad_id: int = -100,
        special_token_label_id: int = 0,
        classifier_dropout: Optional[float] = None,
        freeze_base_model: bool = False,
        warmup_proportion: float = 0.1,
        seq2seq_encoder: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.save_hyperparameters()

        self.special_token_label_id = special_token_label_id

        self.learning_rate = learning_rate
        self.warmup_proportion = warmup_proportion
        self.task_learning_rate = task_learning_rate
        self.label_pad_id = label_pad_id
        self.num_classes = num_classes

        config = AutoConfig.from_pretrained(model_name_or_path)
        if self.is_from_pretrained:
            self.model = AutoModel.from_config(config=config)
        else:
            self.model = AutoModel.from_pretrained(model_name_or_path, config=config)

        if freeze_base_model:
            self.model.requires_grad_(False)

        hidden_size = config.hidden_size
        self.seq2seq_encoder = None
        if seq2seq_encoder is not None:
            self.seq2seq_encoder, hidden_size = build_seq2seq_encoder(
                config=seq2seq_encoder, input_size=hidden_size
            )

        if classifier_dropout is None:
            # Get the classifier dropout value from the Huggingface model config.
            # This is a bit of a mess since some Configs use different variable names or change the semantics
            # of the dropout (e.g. DistilBert has one dropout prob for QA and one for Seq classification, and a
            # general one for embeddings, encoder and pooler).
            classifier_dropout_attr = HF_MODEL_TYPE_TO_CLASSIFIER_DROPOUT_ATTRIBUTE.get(
                config.model_type, "classifier_dropout"
            )
            if hasattr(config, classifier_dropout_attr):
                classifier_dropout = getattr(config, classifier_dropout_attr)
            else:
                raise ValueError(
                    f"The config {type(config),__name__} loaded from {model_name_or_path} has no attribute "
                    f"{classifier_dropout_attr}"
                )
        self.dropout = nn.Dropout(classifier_dropout)

        self.classifier = nn.Linear(hidden_size, num_classes)

        if use_crf:
            try:
                from torchcrf import CRF
            except ImportError:
                raise ImportError(
                    "To use CRFs, the torchcrf package must be installed. "
                    "You can install it with `pip install pytorch-crf`."
                )

            self.crf = CRF(num_tags=num_classes, batch_first=True)
        else:
            self.crf = None

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        result: MutableMapping[str, Union[LongTensor, FloatTensor]] = {}
        logits = outputs.logits
        attention_mask = inputs["attention_mask"]
        special_tokens_mask = inputs["special_tokens_mask"]
        attention_mask_bool = attention_mask.to(torch.bool)
        if self.crf is not None:
            decoded_tags = self.crf.decode(emissions=logits, mask=attention_mask_bool)
            # pad the decoded tags to the length of the logits to have the same shape as when not using the crf
            seq_len = logits.shape[1]
            padded_tags = [
                tags + [self.label_pad_id] * (seq_len - len(tags)) for tags in decoded_tags
            ]
            tags_tensor = torch.tensor(padded_tags, device=logits.device).to(torch.long)
        else:
            # get the max index for each token from the logits
            tags_tensor = torch.argmax(logits, dim=-1).to(torch.long)
        # set the padding and special tokens to the label_pad_id
        mask = attention_mask_bool & ~special_tokens_mask.to(torch.bool)
        tags_tensor = tags_tensor.masked_fill(~mask, self.label_pad_id)
        probs = torch.softmax(logits, dim=-1)

        assert isinstance(tags_tensor, LongTensor)
        assert isinstance(probs, FloatTensor)
        result["labels"] = tags_tensor
        # TODO: is it correct to use this also in the case of the crf?
        result["probabilities"] = probs

        return result

    def forward(
        self, inputs: InputType, targets: Optional[TargetType] = None
    ) -> TokenClassifierOutput:
        inputs_without_special_tokens_mask = {
            k: v for k, v in inputs.items() if k != "special_tokens_mask"
        }
        outputs = self.model(**inputs_without_special_tokens_mask)
        sequence_output = outputs[0]

        if self.seq2seq_encoder is not None:
            sequence_output = self.seq2seq_encoder(sequence_output)

        sequence_output = self.dropout(sequence_output)
        logits = self.classifier(sequence_output)

        loss = None
        if targets is not None:
            labels = targets["labels"]
            if self.crf is not None:
                # Overwrite the padding labels with ignore_index. Note that this is different from the
                # attention_mask, because the attention_mask includes special tokens, whereas the labels
                # are set to label_pad_id also for special tokens (e.g. [CLS]). We need handle all
                # occurrences of label_pad_id because usually that index is out of range with respect to
                # the number of logits in which case the crf would complain. However, we can not simply
                # pass a mask to the crf that also masks out the special tokens, because the crf does not
                # allow the first token to be masked out.
                mask_pad_or_special = labels == self.label_pad_id
                labels_valid = labels.masked_fill(mask_pad_or_special, self.special_token_label_id)
                # the crf expects a bool mask
                if "attention_mask" in inputs:
                    mask_bool = inputs["attention_mask"].to(torch.bool)
                else:
                    mask_bool = None
                log_likelihood = self.crf(emissions=logits, tags=labels_valid, mask=mask_bool)
                loss = -log_likelihood
            else:
                loss_fct = nn.CrossEntropyLoss(ignore_index=self.label_pad_id)
                loss = loss_fct(logits.view(-1, self.num_classes), labels.view(-1))

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def configure_optimizers(self) -> OptimizerLRScheduler:
        if self.task_learning_rate is not None:
            all_params = dict(self.named_parameters())
            base_model_params = dict(self.model.named_parameters(prefix="model"))
            task_params = {k: v for k, v in all_params.items() if k not in base_model_params}
            optimizer = torch.optim.AdamW(
                [
                    {"params": base_model_params.values(), "lr": self.learning_rate},
                    {"params": task_params.values(), "lr": self.task_learning_rate},
                ]
            )
        else:
            optimizer = torch.optim.AdamW(self.parameters(), lr=self.learning_rate)

        if self.warmup_proportion > 0.0:
            stepping_batches = self.trainer.estimated_stepping_batches
            scheduler = get_linear_schedule_with_warmup(
                optimizer, int(stepping_batches * self.warmup_proportion), stepping_batches
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



import copy
import logging
from typing import Any, Dict, Optional, Tuple, Type, Union

import torch
from pie_core import Model
from pie_core.utils.hydra import resolve_type
from pytorch_lightning.utilities.types import OptimizerLRScheduler
from torch import FloatTensor, LongTensor
from torch.optim import Optimizer
from transformers import PreTrainedModel, SchedulerType, get_scheduler
from transformers.modeling_outputs import Seq2SeqLMOutput
from typing_extensions import TypeAlias

from .common import ModelWithBoilerplate

logger = logging.getLogger(__name__)

# model inputs / outputs / targets
InputType: TypeAlias = Dict[str, LongTensor]
OutputType: TypeAlias = Seq2SeqLMOutput
TargetType: TypeAlias = Dict[str, LongTensor]
# step inputs (batch) / outputs (loss)
StepInputType: TypeAlias = Tuple[InputType, TargetType]
StepOutputType: TypeAlias = FloatTensor


@Model.register()
class SimpleGenerativeModel(
    ModelWithBoilerplate[InputType, OutputType, TargetType, StepOutputType],
):
    """This model is a simple wrapper around a generative model from Huggingface transformers. That
    means, its predict() and predict_step() methods will call the generate() method of the base
    model.

    If a taskmodule config is provided, the taskmodule will be instantiated and used to create metrics and
    a generation config with its configure_model_metric() and configure_model_generation() methods,
    respectively.

    If the base model has a configure_optimizer() method, this will be used to create the optimizer. Otherwise,
    the optimizer_type and learning_rate will be used to create an optimizer.

    Args:
        base_model_type: The type of the base model, e.g. "transformers.AutoModelForSeq2SeqLM". It should have a
            from_pretrained() method.
        base_model_config: A dictionary with the keyword arguments that will be passed to the from_pretrained()
            method of the base model.
        override_generation_kwargs: The generation config for the base model. This will override the generation config
            from the taskmodule, if one is provided.
        warmup_proportion: The proportion of the training steps that will be used for the warmup of the learning rate
            scheduler.
        learning_rate: The learning rate for the optimizer. If the base model has a configure_optimizer() method, this
            will be ignored.
        optimizer_type: The type of the optimizer. If the base model has a configure_optimizer() method, this will be
            ignored.
        **kwargs: Additional keyword arguments that will be passed to the PyTorchIEModel constructor.
    """

    def __init__(
        self,
        # base model setup
        base_model: Optional[Dict[str, Any]] = None,
        # old setup
        base_model_type: Optional[str] = None,
        base_model_config: Optional[Dict[str, Any]] = None,
        # generation
        override_generation_kwargs: Optional[Dict[str, Any]] = None,
        # optimizer / schedular
        # important: the following entries (optimizer_type and learning_rate) are only used
        # if the base model does not have a configure_optimizer method!
        optimizer_type: Optional[Union[str, Type[Optimizer]]] = None,
        learning_rate: Optional[float] = None,
        warmup_proportion: float = 0.0,
        scheduler_name: Optional[Union[str, SchedulerType]] = None,
        scheduler_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        if base_model is None:
            if base_model_type is None:
                raise ValueError(
                    "Either base_model or base_model_type must be provided. If base_model is not provided, "
                    "base_model_type must be a valid model type, e.g. 'transformers.AutoModelForSeq2SeqLM'."
                )
            logger.warning(
                "The base_model_type and base_model_config arguments are deprecated. Please use base_model. "
                "You can use the following code to create the base_model argument: "
                "base_model = {'_type_': base_model_type, **base_model_config}"
            )
            base_model = {"_type_": base_model_type, **(base_model_config or {})}

        if scheduler_name is None and warmup_proportion > 0.0:
            logger.warning(
                "warmup_proportion is set to a value > 0.0, but scheduler_name is not set. "
                "Setting scheduler_name to 'linear' by default."
            )
            scheduler_name = "linear"

        self.save_hyperparameters(ignore=["base_model_type", "base_model_config"])

        # optimizer / scheduler
        self.learning_rate = learning_rate
        self.optimizer_type = optimizer_type
        self.scheduler_name = scheduler_name
        self.warmup_proportion = warmup_proportion
        self.scheduler_kwargs = scheduler_kwargs or {}

        self.model = self.setup_base_model(config=base_model)
        self.generation_config = self.configure_generation(**(override_generation_kwargs or {}))

    def setup_base_model(self, config: Dict[str, Any]) -> PreTrainedModel:
        config = copy.copy(config)
        resolved_base_model_type: Type[PreTrainedModel] = resolve_type(config.pop("_type_"))
        return resolved_base_model_type.from_pretrained(**config)

    def configure_generation(self, **kwargs) -> Dict[str, Any]:
        if self.taskmodule is not None:
            # get the generation config from the taskmodule
            generation_config = self.taskmodule.configure_model_generation()
        else:
            logger.warning(
                "No taskmodule is available, so no generation config will be created. Consider "
                "setting taskmodule_config to a valid taskmodule config to use specific setup for generation."
            )
            generation_config = {}
        generation_config.update(kwargs)
        return generation_config

    def predict(self, inputs, **kwargs) -> TargetType:
        is_training = self.training
        self.eval()

        generation_kwargs = copy.deepcopy(self.generation_config)
        generation_kwargs.update(kwargs)
        outputs = self.model.generate(**inputs, **generation_kwargs)

        if is_training:
            self.train()

        # TODO: move into base model? or does this work for "all" generative models?
        # strip the bos_id
        if isinstance(outputs, torch.Tensor):
            labels = outputs[:, 1:]
            assert isinstance(labels, LongTensor)
            return {"labels": labels}
        else:
            raise ValueError(f"Unsupported output type: {type(outputs)}")

    def forward(self, inputs: InputType, targets: Optional[TargetType] = None) -> OutputType:
        kwargs = {**inputs, **(targets or {})}
        return self.model(**kwargs)

    def decode(self, inputs: InputType, outputs: OutputType) -> TargetType:
        # construct prediction from the model output
        logits = outputs.logits
        # get the indices (these are without the initial bos_ids, see above)
        prediction = torch.argmax(logits, dim=-1).to(torch.long)
        assert isinstance(prediction, LongTensor)
        return {"labels": prediction}

    def configure_optimizers(self) -> OptimizerLRScheduler:
        if hasattr(self.model, "configure_optimizer") and callable(self.model.configure_optimizer):
            if self.learning_rate is not None:
                raise ValueError(
                    f"learning_rate is set to {self.learning_rate}, but the *base model* ({type(self.model)}) has a "
                    f"configure_optimizer method. Please set learning_rate to None and configure the optimizer "
                    f"inside the *base model*."
                )
            optimizer = self.model.configure_optimizer()
        else:
            logger.warning(
                f"The model does not have a configure_optimizer method. Creating an optimizer of "
                f"optimizer_type={self.optimizer_type} with the learning_rate={self.learning_rate} instead."
            )
            if self.optimizer_type is None:
                raise ValueError(
                    f"optimizer_type is None, but the *base model* ({type(self.model)}) does not have a "
                    f"configure_optimizer method. Please set the optimizer_type to a valid optimizer type, "
                    f"e.g. optimizer_type=torch.optim.Adam."
                )
            resolved_optimizer_type = resolve_type(
                self.optimizer_type, expected_super_type=Optimizer
            )
            optimizer = resolved_optimizer_type(self.parameters(), lr=self.learning_rate)

        if self.scheduler_name is not None:
            num_training_steps = self.trainer.estimated_stepping_batches
            num_warmup_steps = int(num_training_steps * self.warmup_proportion)
            scheduler = get_scheduler(
                name=self.scheduler_name,
                optimizer=optimizer,
                num_warmup_steps=num_warmup_steps,
                num_training_steps=num_training_steps,
                scheduler_specific_kwargs=self.scheduler_kwargs,
            )
            return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        else:
            return optimizer



import dataclasses
import logging
from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple, TypeVar

import pandas as pd
import torch
import torch.nn.functional as F
from pie_core import Annotation
from torch import Tensor

logger = logging.getLogger(__name__)


def _pad_tensor(tensor: Tensor, target_shape: List[int], pad_value: float) -> Tensor:
    """Pad a tensor to a target shape.

    Args:
        tensor: The tensor to pad.
        target_shape: The target shape.
        pad_value: The value to use for padding.

    Returns: The padded tensor.
    """

    shape = tensor.shape
    pad: List[int] = []
    for i, s in enumerate(shape):
        pad = [0, target_shape[i] - s] + pad
    result = F.pad(tensor, pad=pad, value=pad_value)

    return result


def maybe_pad_values(
    values: Any, pad_value: Optional[Any] = None, strategy: str = "longest"
) -> Any:
    """If an iterable of values is passed and a pad value is given, pad the values to the same
    length and create a tensor from them. Otherwise, return the values unchanged.

    Note that the padding is done on all dimensions.

    Args:
        values: The values to pad.
        pad_value: The value to use for padding.
        strategy: The padding strategy. Currently only "longest" is supported.

    Returns: The padded values.
    """

    if pad_value is None:
        return values
    if not isinstance(values, Iterable):
        raise TypeError(f"values must be iterable to pad them, but got {type(values)}")
    if strategy != "longest":
        raise ValueError(f"unknown padding strategy: {strategy}")
    tensor_list = [torch.tensor(value_list) for value_list in values]
    shape_lists = list(zip(*[t.shape for t in tensor_list]))
    max_shape = [max(dims) for dims in shape_lists]
    padded = [
        _pad_tensor(tensor=t, target_shape=max_shape, pad_value=pad_value)
        for i, t in enumerate(tensor_list)
    ]
    return torch.stack(padded)


def maybe_to_tensor(
    values: Iterable[Any], dtype: Optional[torch.dtype] = None, pad_value: Optional[Any] = None
) -> Any:
    """If an iterable of values is passed and a dtype is given, convert the values to a tensor of
    the given type.

    Args:
        values: The values to convert.
        dtype: A dtype to convert the values to.
        pad_value: A pad value to use if the values are padded.

    Returns: A tensor or the values unchanged.
    """

    if all(v is None for v in values):
        return None
    if dtype is None:
        return values
    maybe_padded = maybe_pad_values(values=values, pad_value=pad_value)
    if not isinstance(maybe_padded, torch.Tensor):
        maybe_padded = torch.Tensor(maybe_padded)
    tensor = maybe_padded.to(dtype=dtype)
    return tensor


class BatchableMixin:
    """A mixin class that provides a batch method to batch a list of instances of the class. All
    attributes, but also property methods, are batched. The batch method returns a dictionary with
    all attribute / property names as keys. The values are tensors created from the stacked values
    of the attributes / properties. The tensors are padded to the length of the longest instance in
    the batch and converted to the given dtype.

    Example:
        >>> import dataclasses
        >>> from typing import List, Dict
        >>> import torch
        >>>
        >>> @dataclasses.dataclass
        >>> class Foo(BatchableMixin):
        >>>     a: List[int]
        >>>
        >>>   @property
        >>>   def len_a(self):
        >>>       return len(self.a)
        >>>
        >>> x = Foo(a=[1, 2, 3])
        >>> y = Foo(a=[4, 5])
        >>>
        >>> Foo.batch(values=[x, y], dtypes={"a": torch.int64, "len_a": torch.int64}, pad_values={"a": 0})
        {'a': tensor([[1, 2, 3],[4, 5, 0]]), 'len_a': tensor([3, 2])}
    """

    @classmethod
    def get_property_names(cls) -> List[str]:
        return [name for name in cls.__dict__ if isinstance(getattr(cls, name), property)]

    @classmethod
    def get_dataclass_field_names(cls) -> List[str]:
        if dataclasses.is_dataclass(cls):
            return [f.name for f in dataclasses.fields(cls)]
        else:
            return []

    @classmethod
    def get_attribute_names(cls) -> List[str]:
        return cls.get_property_names() + cls.get_dataclass_field_names()

    @classmethod
    def batch(
        cls,
        values: List[Any],
        dtypes: Dict[str, torch.dtype],
        pad_values: Dict[str, Any],
    ) -> Dict[str, Any]:
        attribute_names = cls.get_attribute_names()
        return {
            k: maybe_to_tensor(
                values=[getattr(x, k) for x in values],
                dtype=dtypes.get(k, None),
                pad_value=pad_values.get(k, None),
            )
            for k in attribute_names
            # Only batch attributes that are not None for any of the values.
            if not all(getattr(x, k) is None for x in values)
        }


T = TypeVar("T")


def increase_counter(
    key: Tuple[Any, ...],
    statistics: Dict[Tuple[Any, ...], int],
    value: int = 1,
):
    key_s = tuple(str(k) for k in key)
    statistics[key_s] += value


class StatisticsMixin(ABC, Generic[T]):
    """A mixin class that provides methods to collect and format statistics.

    Args:
        collect_statistics: Control whether statistics should be collected.
            If `False`, the mixin will not show any statistics when calling
            `show_statistics`. Further effects depend on the implementation
            of the mixin.
        **kwargs: Additional keyword arguments to pass to the parent class.
    """

    def __init__(self, collect_statistics: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.collect_statistics = collect_statistics
        self.reset_statistics()

    @abstractmethod
    def reset_statistics(self):
        """Reset the statistics collected by this mixin (state)."""
        pass

    @abstractmethod
    def get_statistics(self) -> T:
        """Get the statistics collected by this mixin.

        This should *not* modify the state of the mixin, repeated calls should return the same
        result!
        """
        pass

    def format_statistics(self, statistics: T) -> str:
        """Format the statistics collected by this mixin as string for display (usually on
        console)."""
        raise NotImplementedError(
            f"format_statistics is not implemented for {self.__class__.__name__}. "
            "Please implement this method to show formatted statistics."
        )

    def show_statistics(self):
        if self.collect_statistics:
            logger.info(f"statistics:\n{self.format_statistics(self.get_statistics())}")


class RelationStatisticsMixin(StatisticsMixin[Dict[Tuple[str, str], int]]):
    """A mixin class that provides methods to collect and format statistics about relations.

    This mixin collects statistics about relations, such as the number of available, used, and
    skipped relations.
    """

    def get_none_label_for_statistics(self) -> str:
        if not hasattr(self, "_statistics_none_label"):
            if hasattr(self, "none_label"):
                # If the mixin has a `none_label` attribute, use it as the label for "no relation".
                self._statistics_none_label = self.none_label
            else:
                self._statistics_none_label = "no_relation"
                logger.warning(
                    f"{type(self).__name__} does not have a `none_label` attribute. "
                    "Using default value 'no_relation'. "
                    "`none_label` is used as the label for relations with score 0 in statistics and "
                    "all relations with label different from `none_label` will be summarized to 'all_relations'. "
                    "Set the `none_label` attribute before using statistics or "
                    "overwrite `get_none_label_for_statistics()` function to get rid of this message."
                )

        return self._statistics_none_label

    def reset_statistics(self):
        self._collected_relations: Dict[str, List[Annotation]] = defaultdict(list)

    def collect_relation(self, kind: str, relation: Annotation):
        if self.collect_statistics:
            self._collected_relations[kind].append(relation)

    def collect_all_relations(self, kind: str, relations: Iterable[Annotation]):
        if self.collect_statistics:
            self._collected_relations[kind].extend(relations)

    def get_statistics(self) -> Dict[Tuple[str, str], int]:
        if self.collect_statistics:
            # create statistics from the collected relations
            statistics: Dict[Tuple[str, str], int] = defaultdict(int)
            all_relations = set(self._collected_relations["available"])
            used_relations = set(self._collected_relations["used"])
            skipped_other = all_relations - used_relations
            for key, rels in self._collected_relations.items():
                rels_set = set(rels)
                if key.startswith("skipped_"):
                    skipped_other -= rels_set
                elif key.startswith("used_"):
                    pass
                elif key in ["available", "used"]:
                    pass
                else:
                    raise ValueError(f"unknown key: {key}")
                for rel in rels_set:
                    # Set `none_label` as label when the score is zero. We encode negative relations
                    # in such a way in the case of multi-label or binary (similarity for coref).
                    label = rel.label if rel.score > 0 else self.get_none_label_for_statistics()
                    increase_counter(key=(key, label), statistics=statistics)
            for rel in skipped_other:
                increase_counter(key=("skipped_other", rel.label), statistics=statistics)

            return dict(statistics)
        else:
            return {}

    def format_statistics(self, statistics: Dict[Tuple[str, str], int]) -> str:
        if len(statistics) > 0:
            to_show_series = pd.Series(statistics)
            # unstack index to have relation labels as column names
            to_show = to_show_series.unstack()
        else:
            # If there were no statistics, create an empty dummy dataframe.
            to_show = pd.DataFrame(pd.Series(dict()))
        # fill missing values with 0 and convert back to int (unstacking may introduce NaNs which are float type)
        to_show = to_show.fillna(0).astype(int)
        if to_show.columns.size > 1:
            to_show["all_relations"] = to_show.loc[
                :, to_show.columns != self.get_none_label_for_statistics()
            ].sum(axis=1)

        #  transpose
        #  to have the labels (which may be a lot) as index for improved readability and
        #  to allow to keep counts as int columns (dtypes are per-column, not per-row)
        to_show = to_show.T
        if "used" in to_show.columns and "available" in to_show.columns:
            to_show["used %"] = (100 * to_show["used"] / to_show["available"]).round()

        return to_show.to_markdown()



import logging
from typing import Any, Callable, Dict, Tuple, Union

import torch
from torch import Tensor, cat, nn

# possible pooler types
CLS_TOKEN = "cls_token"  # CLS token
START_TOKENS = "start_tokens"  # MTB start tokens concat
MENTION_POOLING = "mention_pooling"  # mention token pooling and concat


logger = logging.getLogger(__name__)


def pool_cls(hidden_state: Tensor, **kwargs) -> Tensor:
    return hidden_state[:, 0, :]


class AtIndexPooler(nn.Module):
    """Pooler that takes the hidden states at given indices. If the index is negative, a learned
    embedding is used.

    The indices are expected to have the shape [batch_size, num_indices]. The resulting embeddings are concatenated,
    so the output shape is [batch_size, num_indices * input_dim].

    Args:
        input_dim: The input dimension of the hidden state.
        num_indices: The number of indices to pool.
        offset: An offset to add to the indices. This can be useful if the input is prepared with special
            tokens at the beginning / at the end of indexed sequences, and we want to use the hidden state of this
            token instead of the first / last token of the sequence.

    Returns:
        The pooled hidden states with shape [batch_size, num_indices * input_dim].
    """

    def __init__(self, input_dim: int, num_indices: int = 2, offset: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.num_indices = num_indices
        self.offset = offset
        self.missing_embeddings = nn.Parameter(torch.empty(num_indices, self.input_dim))
        nn.init.normal_(self.missing_embeddings)

    def forward(self, hidden_state: Tensor, indices: Tensor, **kwargs) -> Tensor:
        batch_size, seq_len, hidden_size = hidden_state.shape
        if indices.shape[1] != self.num_indices:
            raise ValueError(
                f"number of indices [{indices.shape[1]}] has to be the same as num_types [{self.num_indices}]"
            )

        # respect the offset
        indices = indices + self.offset

        # times num_types due to concat
        result = torch.zeros(
            batch_size, hidden_size * self.num_indices, device=hidden_state.device
        )
        for batch_idx, current_indices in enumerate(indices):
            current_embeddings = [
                (
                    hidden_state[batch_idx, current_indices[i], :]
                    if current_indices[i] >= 0
                    else self.missing_embeddings[i]
                )
                for i in range(self.num_indices)
            ]
            result[batch_idx] = cat(current_embeddings, 0)
        return result

    @property
    def output_dim(self) -> int:
        return self.input_dim * self.num_indices


class ArgumentWrappedPooler(nn.Module):
    """Wraps a pooler and maps the arguments to the pooler.

    Args:
        pooler: The pooler to wrap.
        argument_mapping: A mapping from the arguments of the forward method to the arguments of the pooler.
    """

    def __init__(
        self, pooler: Union[nn.Module, Callable], argument_mapping: Dict[str, str], **kwargs
    ):
        super().__init__(**kwargs)
        self.pooler = pooler
        self.argument_mapping = argument_mapping

    def forward(self, hidden_state: Tensor, **kwargs) -> Tensor:
        pooler_kwargs = {}
        for k, v in kwargs.items():
            if k in self.argument_mapping:
                pooler_kwargs[self.argument_mapping[k]] = v
        return self.pooler(hidden_state, **pooler_kwargs)


class SpanMaxPooler(nn.Module):
    """Pooler that takes the max hidden state over spans. If the start or end index is negative, a
    learned.

    embedding is used. The indices are expected to have the shape [batch_size, num_indices]. The resulting embeddings
    are concatenated, so the output shape is [batch_size, num_indices * input_dim].

    Args:
        input_dim: The input dimension of the hidden state.
        num_indices: The number of indices to pool.

    Returns:
        The pooled hidden states with shape [batch_size, num_indices * input_dim].
    """

    def __init__(self, input_dim: int, num_indices: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.num_indices = num_indices
        self.missing_embeddings = nn.Parameter(torch.empty(num_indices, self.input_dim))
        nn.init.normal_(self.missing_embeddings)

    def forward(
        self, hidden_state: Tensor, start_indices: Tensor, end_indices: Tensor, **kwargs
    ) -> Tensor:
        batch_size, seq_len, hidden_size = hidden_state.shape
        if start_indices.shape[1] != self.num_indices:
            raise ValueError(
                f"number of start indices [{start_indices.shape[1]}] has to be the same as num_types [{self.num_indices}]"
            )

        if end_indices.shape[1] != self.num_indices:
            raise ValueError(
                f"number of end indices [{end_indices.shape[1]}] has to be the same as num_types [{self.num_indices}]"
            )

        # check that start_indices are before end_indices
        mask_both_positive = (start_indices >= 0) & (end_indices >= 0)
        mask_start_before_end = start_indices < end_indices
        mask_valid = mask_start_before_end | ~mask_both_positive
        if not torch.all(mask_valid):
            raise ValueError(
                f"values in start_indices have to be smaller than respective values in "
                f"end_indices, but start_indices=\n{start_indices}\n and end_indices=\n{end_indices}"
            )

        # times num_indices due to concat
        result = torch.zeros(
            batch_size, hidden_size * self.num_indices, device=hidden_state.device
        )
        for batch_idx in range(batch_size):
            current_start_indices = start_indices[batch_idx]
            current_end_indices = end_indices[batch_idx]
            current_embeddings = [
                (
                    torch.amax(
                        hidden_state[
                            batch_idx, current_start_indices[i] : current_end_indices[i], :
                        ],
                        0,
                    )
                    if current_start_indices[i] >= 0 and current_end_indices[i] >= 0
                    else self.missing_embeddings[i]
                )
                for i in range(self.num_indices)
            ]
            result[batch_idx] = cat(current_embeddings, 0)

        return result

    @property
    def output_dim(self) -> int:
        return self.input_dim * self.num_indices


class SpanMeanPooler(nn.Module):
    """Pooler that takes the mean hidden state over spans. If the start or end index is negative, a
    learned embedding is used. The indices are expected to have the shape [batch_size,
    num_indices].

    The resulting embeddings are concatenated, so the output shape is [batch_size, num_indices * input_dim].
    Note this a slightly modified version of the pytorch_ie.models.components.pooler.SpanMaxPooler,
    i.e. we changed the aggregation method from torch.amax to torch.mean.

    Args:
        input_dim: The input dimension of the hidden state.
        num_indices: The number of indices to pool.

    Returns:
        The pooled hidden states with shape [batch_size, num_indices * input_dim].
    """

    def __init__(self, input_dim: int, num_indices: int = 2, **kwargs):
        super().__init__(**kwargs)
        self.input_dim = input_dim
        self.num_indices = num_indices
        self.missing_embeddings = nn.Parameter(torch.empty(num_indices, self.input_dim))
        nn.init.normal_(self.missing_embeddings)

    def forward(
        self, hidden_state: Tensor, start_indices: Tensor, end_indices: Tensor, **kwargs
    ) -> Tensor:
        batch_size, seq_len, hidden_size = hidden_state.shape
        if start_indices.shape[1] != self.num_indices:
            raise ValueError(
                f"number of start indices [{start_indices.shape[1]}] has to be the same as num_types [{self.num_indices}]"
            )

        if end_indices.shape[1] != self.num_indices:
            raise ValueError(
                f"number of end indices [{end_indices.shape[1]}] has to be the same as num_types [{self.num_indices}]"
            )

        # check that start_indices are before end_indices
        mask_both_positive = (start_indices >= 0) & (end_indices >= 0)
        mask_start_before_end = start_indices < end_indices
        mask_valid = mask_start_before_end | ~mask_both_positive
        if not torch.all(mask_valid):
            raise ValueError(
                f"values in start_indices have to be smaller than respective values in "
                f"end_indices, but start_indices=\n{start_indices}\n and end_indices=\n{end_indices}"
            )

        # times num_indices due to concat
        result = torch.zeros(
            batch_size, hidden_size * self.num_indices, device=hidden_state.device
        )
        for batch_idx in range(batch_size):
            current_start_indices = start_indices[batch_idx]
            current_end_indices = end_indices[batch_idx]
            current_embeddings = [
                (
                    torch.mean(
                        hidden_state[
                            batch_idx, current_start_indices[i] : current_end_indices[i], :
                        ],
                        dim=0,
                    )
                    if current_start_indices[i] >= 0 and current_end_indices[i] >= 0
                    else self.missing_embeddings[i]
                )
                for i in range(self.num_indices)
            ]
            result[batch_idx] = cat(current_embeddings, 0)

        return result

    @property
    def output_dim(self) -> int:
        return self.input_dim * self.num_indices


def get_pooler_and_output_size(config: Dict[str, Any], input_dim: int) -> Tuple[Callable, int]:
    pooler_config = dict(config)
    pooler_type = pooler_config.pop("type", CLS_TOKEN)
    pooler: Callable
    if pooler_type == CLS_TOKEN:
        return pool_cls, input_dim
    elif pooler_type == START_TOKENS:
        pooler = AtIndexPooler(input_dim=input_dim, offset=-1, **pooler_config)
        pooler_wrapped = ArgumentWrappedPooler(
            pooler=pooler, argument_mapping={"start_indices": "indices"}
        )
        return pooler_wrapped, pooler.output_dim
    elif pooler_type == MENTION_POOLING:
        aggregate = pooler_config.pop("aggregate", "max")
        if aggregate == "max":
            pooler = SpanMaxPooler(input_dim=input_dim, **pooler_config)
            return pooler, pooler.output_dim
        elif aggregate == "mean":
            pooler = SpanMeanPooler(input_dim=input_dim, **pooler_config)
            return pooler, pooler.output_dim
        else:
            raise ValueError(f'Unknown aggregation method for mention pooling: "{aggregate}"')
    else:
        raise ValueError(f'Unknown pooler type "{pooler_type}"')



import math
from typing import Callable, List

import torch
from transformers import LogitsProcessor, add_start_docstrings
from transformers.generation.logits_process import LOGITS_PROCESSOR_INPUTS_DOCSTRING


class PrefixConstrainedLogitsProcessorWithMaximum(LogitsProcessor):
    r"""This is similar to [`PrefixConstrainedLogitsProcessor`] but the constraint function gets the
    maximum possible index as input. This is useful for Pointer Network where the generated token
    can be an index into the input which depends on the length of that input.

    Args:
        prefix_allowed_tokens_fn (Callable[[int, torch.LongTensor, int], List[int]]):
            Should return the list of token ids allowed at the next generation step,
            given (`batch_id`, `input_ids_so_far`, `max_index`).
    """

    def __init__(
        self,
        prefix_allowed_tokens_fn: Callable[[int, torch.LongTensor, int], List[int]],
        num_beams: int,
    ):
        self._prefix_allowed_tokens_fn = prefix_allowed_tokens_fn
        self._num_beams = num_beams

    @add_start_docstrings(LOGITS_PROCESSOR_INPUTS_DOCSTRING)
    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        if not torch.isfinite(scores).all():
            raise ValueError(
                "scores contains ±inf or NaN, which is not allowed by "
                "PrefixConstrainedLogitsProcessorWithMaximum. "
                "Insert FinitizeLogitsProcessor earlier to clean them."
            )
        mask = torch.full_like(scores, -math.inf)
        for batch_id, beam_sent in enumerate(
            input_ids.view(-1, self._num_beams, input_ids.shape[-1])
        ):
            for beam_id, sent in enumerate(beam_sent):
                allowed_ids = self._prefix_allowed_tokens_fn(batch_id, sent, mask.size(1))
                if len(allowed_ids) == 0:
                    raise ValueError(
                        f"No allowed token ids for batch_id {batch_id}, beam_id {beam_id} with "
                        f"previous ids: {sent}. This would result in undefined behaviour, "
                        "so this is not allowed. Please adjust the prefix_allowed_tokens_fn "
                        "implementation."
                    )
                mask[batch_id * self._num_beams + beam_id, allowed_ids] = 0

        result = scores + mask
        assert isinstance(result, torch.FloatTensor)
        return result


class FinitizeLogitsProcessor(LogitsProcessor):
    r"""Replaces any `±inf` logits with the largest-magnitude finite values for the tensor’s dtype,
    ensuring all logits are valid for downstream ops."""

    @add_start_docstrings(LOGITS_PROCESSOR_INPUTS_DOCSTRING)
    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor
    ) -> torch.FloatTensor:
        finite_min = torch.finfo(scores.dtype).min
        finite_max = torch.finfo(scores.dtype).max
        # Use nan_to_num for a fast, fused replacement (PyTorch ≥ 1.8)
        result = torch.nan_to_num(scores, neginf=finite_min, posinf=finite_max)
        assert isinstance(result, torch.FloatTensor)
        return result



# Copyright 2021 The Fairseq Authors and The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch BART model, but the decoder accepts predefined position ids. If not provided, the
original logic is used to create the position ids.

The model is based on the BartModel from Transformers 4.35.0,
i.e. https://github.com/huggingface/transformers/blob/v4.35.0/src/transformers/models/bart/modeling_bart.py.

Note: This also contains some minor modifications to make the code mypy (v1.4.1) compliant.
.
"""
import math
from typing import Any, List, Optional, Tuple, Union

import torch
import torch.utils.checkpoint
from torch import nn
from transformers.modeling_attn_mask_utils import (
    _prepare_4d_attention_mask,
    _prepare_4d_causal_attention_mask,
)
from transformers.modeling_outputs import (
    BaseModelOutput,
    BaseModelOutputWithPastAndCrossAttentions,
    Seq2SeqModelOutput,
)
from transformers.models.bart import BartConfig
from transformers.models.bart.modeling_bart import (
    _CHECKPOINT_FOR_DOC,
    _CONFIG_FOR_DOC,
    _EXPECTED_OUTPUT_SHAPE,
    BART_INPUTS_DOCSTRING,
    BART_START_DOCSTRING,
    BartDecoderLayer,
    BartEncoder,
    BartPreTrainedModel,
    shift_tokens_right,
)
from transformers.utils import (
    add_code_sample_docstrings,
    add_start_docstrings,
    add_start_docstrings_to_model_forward,
    logging,
)

logger = logging.get_logger(__name__)


class BartLearnedPositionalEmbeddingWithPositionIds(nn.Embedding):
    """This module learns positional embeddings up to a fixed maximum size."""

    def __init__(self, num_embeddings: int, embedding_dim: int):
        # Bart is set up so that if padding_idx is specified then offset the embedding ids by 2
        # and adjust num_embeddings appropriately. Other models don't have this hack
        self.offset = 2
        super().__init__(num_embeddings + self.offset, embedding_dim)

    def forward(
        self,
        input_ids: torch.Tensor,
        past_key_values_length: int = 0,
        position_ids: Optional[torch.Tensor] = None,
    ):
        """`input_ids' shape is expected to be [bsz x seqlen]."""

        if position_ids is None:
            bsz, seq_len = input_ids.shape[:2]
            positions = torch.arange(
                past_key_values_length,
                past_key_values_length + seq_len,
                dtype=torch.long,
                device=self.weight.device,
            ).expand(bsz, -1)
        else:
            positions = position_ids

        return super().forward(positions + self.offset)


class BartDecoderWithPositionIds(BartPreTrainedModel):
    """Transformer decoder consisting of *config.decoder_layers* layers. Each layer is a
    [`BartDecoderLayer`]

    Args:
        config: BartConfig
        embed_tokens (nn.Embedding): output embedding
    """

    def __init__(self, config: BartConfig, embed_tokens: Optional[nn.Embedding] = None):
        super().__init__(config)
        self.dropout = config.dropout
        self.layerdrop = config.decoder_layerdrop
        self.padding_idx = config.pad_token_id
        self.max_target_positions = config.max_position_embeddings
        self.embed_scale = math.sqrt(config.d_model) if config.scale_embedding else 1.0

        self.embed_tokens = nn.Embedding(config.vocab_size, config.d_model, self.padding_idx)

        if embed_tokens is not None:
            self.embed_tokens.weight = embed_tokens.weight

        self.embed_positions = BartLearnedPositionalEmbeddingWithPositionIds(
            config.max_position_embeddings,
            config.d_model,
        )
        self.layers = nn.ModuleList(
            [BartDecoderLayer(config) for _ in range(config.decoder_layers)]
        )
        self.layernorm_embedding = nn.LayerNorm(config.d_model)

        self.gradient_checkpointing = False
        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embed_tokens

    def set_input_embeddings(self, value):
        self.embed_tokens = value

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.LongTensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, BaseModelOutputWithPastAndCrossAttentions]:
        r"""
        Args:
            input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
                Indices of input sequence tokens in the vocabulary. Padding will be ignored by default should you
                provide it.

                Indices can be obtained using [`AutoTokenizer`]. See [`PreTrainedTokenizer.encode`] and
                [`PreTrainedTokenizer.__call__`] for details.

                [What are input IDs?](../glossary#input-ids)
            position_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`, *optional*):
                Position indices for each input sequence token that are used to create the position embedding
                of the sequence. If `None` (default), position ids are automatically created as sequential
                integers (takes previous `past_key_values` into account, if provided).
            attention_mask (`torch.Tensor` of shape `(batch_size, sequence_length)`, *optional*):
                Mask to avoid performing attention on padding token indices. Mask values selected in `[0, 1]`:

                - 1 for tokens that are **not masked**,
                - 0 for tokens that are **masked**.

                [What are attention masks?](../glossary#attention-mask)
            encoder_hidden_states (`torch.FloatTensor` of shape `(batch_size, encoder_sequence_length, hidden_size)`, *optional*):
                Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention
                of the decoder.
            encoder_attention_mask (`torch.LongTensor` of shape `(batch_size, encoder_sequence_length)`, *optional*):
                Mask to avoid performing cross-attention on padding tokens indices of encoder input_ids. Mask values
                selected in `[0, 1]`:

                - 1 for tokens that are **not masked**,
                - 0 for tokens that are **masked**.

                [What are attention masks?](../glossary#attention-mask)
            head_mask (`torch.Tensor` of shape `(decoder_layers, decoder_attention_heads)`, *optional*):
                Mask to nullify selected heads of the attention modules. Mask values selected in `[0, 1]`:

                - 1 indicates the head is **not masked**,
                - 0 indicates the head is **masked**.

            cross_attn_head_mask (`torch.Tensor` of shape `(decoder_layers, decoder_attention_heads)`, *optional*):
                Mask to nullify selected heads of the cross-attention modules in the decoder to avoid performing
                cross-attention on hidden heads. Mask values selected in `[0, 1]`:

                - 1 indicates the head is **not masked**,
                - 0 indicates the head is **masked**.

            past_key_values (`tuple(tuple(torch.FloatTensor))`, *optional*, returned when `use_cache=True` is passed or when `config.use_cache=True`):
                Tuple of `tuple(torch.FloatTensor)` of length `config.n_layers`, with each tuple having 2 tensors of
                shape `(batch_size, num_heads, sequence_length, embed_size_per_head)`) and 2 additional tensors of
                shape `(batch_size, num_heads, encoder_sequence_length, embed_size_per_head)`.

                Contains pre-computed hidden-states (key and values in the self-attention blocks and in the
                cross-attention blocks) that can be used (see `past_key_values` input) to speed up sequential decoding.

                If `past_key_values` are used, the user can optionally input only the last `decoder_input_ids` (those
                that don't have their past key value states given to this model) of shape `(batch_size, 1)` instead of
                all `decoder_input_ids` of shape `(batch_size, sequence_length)`.
            inputs_embeds (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`, *optional*):
                Optionally, instead of passing `input_ids` you can choose to directly pass an embedded
                representation. This is useful if you want more control over how to convert `input_ids` indices
                into associated vectors than the model's internal embedding lookup matrix.
            output_attentions (`bool`, *optional*):
                Whether or not to return the attentions tensors of all attention layers. See `attentions` under
                returned tensors for more detail.
            output_hidden_states (`bool`, *optional*):
                Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
                for more detail.
            return_dict (`bool`, *optional*):
                Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        """
        output_attentions = (
            output_attentions if output_attentions is not None else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        # retrieve input_ids and inputs_embeds
        if input_ids is not None and inputs_embeds is not None:
            raise ValueError(
                "You cannot specify both decoder_input_ids and decoder_inputs_embeds at the same time"
            )
        elif input_ids is not None:
            # input = input_ids
            input_shape = input_ids.shape
            # TODO: check if commenting this out is really fine
            # input_ids = input_ids.view(-1, input_shape[-1])
            inputs_embeds = self.embed_tokens(input_ids) * self.embed_scale
        elif inputs_embeds is not None:
            # (bsz, seq_len)
            input_shape = inputs_embeds.size()[:-1]
            # input = inputs_embeds[:, :, -1]
            # create dummy input_ids with correct shape for position embedding
            dummy_input_ids = torch.zeros(
                input_shape, dtype=torch.long, device=inputs_embeds.device
            )
            assert isinstance(dummy_input_ids, torch.LongTensor)
            input_ids = dummy_input_ids

        if input_ids is None or inputs_embeds is None:
            raise ValueError(
                "You have to specify either decoder_input_ids or decoder_inputs_embeds"
            )

        # past_key_values_length
        past_key_values_length = (
            past_key_values[0][0].shape[2] if past_key_values is not None else 0
        )

        # if inputs_embeds is None:
        #    inputs_embeds = self.embed_tokens(input) * self.embed_scale

        if getattr(self.config, "_flash_attn_2_enabled", False):
            # 2d mask is passed through the layers
            attention_mask = (
                attention_mask if (attention_mask is not None and 0 in attention_mask) else None
            )
        else:
            # 4d mask is passed through the layers
            attention_mask = _prepare_4d_causal_attention_mask(
                attention_mask, input_shape, inputs_embeds, past_key_values_length
            )

        # expand encoder attention mask
        if encoder_hidden_states is not None and encoder_attention_mask is not None:
            if getattr(self.config, "_flash_attn_2_enabled", False):
                encoder_attention_mask = (
                    encoder_attention_mask if 0 in encoder_attention_mask else None
                )
            else:
                # [bsz, seq_len] -> [bsz, 1, tgt_seq_len, src_seq_len]
                encoder_attention_mask = _prepare_4d_attention_mask(
                    encoder_attention_mask, inputs_embeds.dtype, tgt_len=input_shape[-1]
                )

        # embed positions
        if position_ids is not None and position_ids.shape != input_shape:
            raise ValueError(
                f"Position IDs shape {position_ids.shape} does not match input ids shape {input_shape}."
            )
        positions = self.embed_positions(input_ids, past_key_values_length, position_ids)
        positions = positions.to(inputs_embeds.device)

        hidden_states = inputs_embeds + positions
        hidden_states = self.layernorm_embedding(hidden_states)

        hidden_states = nn.functional.dropout(
            hidden_states, p=self.dropout, training=self.training
        )

        if self.gradient_checkpointing and self.training:
            if use_cache:
                logger.warning_once(
                    "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                )
                use_cache = False

        # decoder layers
        all_hidden_states: Optional[Tuple[Any, ...]] = () if output_hidden_states else None
        all_self_attns: Optional[Tuple[Any, ...]] = () if output_attentions else None
        all_cross_attentions: Optional[Tuple[Any, ...]] = (
            () if (output_attentions and encoder_hidden_states is not None) else None
        )
        next_decoder_cache: Optional[Tuple[Any, ...]] = () if use_cache else None

        # check if head_mask/cross_attn_head_mask has a correct number of layers specified if desired
        for attn_mask, mask_name in zip(
            [head_mask, cross_attn_head_mask], ["head_mask", "cross_attn_head_mask"]
        ):
            if attn_mask is not None:
                if attn_mask.size()[0] != (len(self.layers)):
                    raise ValueError(
                        f"The `{mask_name}` should be specified for {len(self.layers)} layers, but it is for"
                        f" {attn_mask.size()[0]}."
                    )

        for idx, decoder_layer in enumerate(self.layers):
            # add LayerDrop (see https://arxiv.org/abs/1909.11556 for description)
            if all_hidden_states is not None:
                all_hidden_states += (hidden_states,)
            if self.training:
                dropout_probability = torch.rand([])
                if dropout_probability < self.layerdrop:
                    continue

            past_key_value = past_key_values[idx] if past_key_values is not None else None

            if self.gradient_checkpointing and self.training:
                layer_outputs = self._gradient_checkpointing_func(
                    decoder_layer.__call__,
                    hidden_states,
                    attention_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    head_mask[idx] if head_mask is not None else None,
                    cross_attn_head_mask[idx] if cross_attn_head_mask is not None else None,
                    None,
                    output_attentions,
                    use_cache,
                )
            else:
                layer_outputs = decoder_layer(
                    hidden_states,
                    attention_mask=attention_mask,
                    encoder_hidden_states=encoder_hidden_states,
                    encoder_attention_mask=encoder_attention_mask,
                    layer_head_mask=(head_mask[idx] if head_mask is not None else None),
                    cross_attn_layer_head_mask=(
                        cross_attn_head_mask[idx] if cross_attn_head_mask is not None else None
                    ),
                    past_key_value=past_key_value,
                    output_attentions=output_attentions,
                    use_cache=use_cache,
                )
            hidden_states = layer_outputs[0]

            if next_decoder_cache is not None:
                next_decoder_cache += (layer_outputs[3 if output_attentions else 1],)

            if all_self_attns is not None:
                all_self_attns += (layer_outputs[1],)

                if all_cross_attentions is not None:
                    all_cross_attentions += (layer_outputs[2],)

        # add hidden states from the last decoder layer
        if all_hidden_states is not None:
            all_hidden_states += (hidden_states,)

        next_cache = next_decoder_cache if use_cache else None
        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    next_cache,
                    all_hidden_states,
                    all_self_attns,
                    all_cross_attentions,
                ]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attns,
            cross_attentions=all_cross_attentions,
        )


@add_start_docstrings(
    "The bare BART Model outputting raw hidden-states without any specific head on top.",
    BART_START_DOCSTRING,
)
class BartModelWithDecoderPositionIds(BartPreTrainedModel):
    _tied_weights_keys = ["encoder.embed_tokens.weight", "decoder.embed_tokens.weight"]

    def __init__(self, config: BartConfig):
        super().__init__(config)

        padding_idx, vocab_size = config.pad_token_id, config.vocab_size
        self.shared = nn.Embedding(vocab_size, config.d_model, padding_idx)

        self.encoder = BartEncoder(config, self.shared)
        self.decoder = BartDecoderWithPositionIds(config, self.shared)

        # Initialize weights and apply final processing
        self.post_init()

    def _tie_weights(self):
        if self.config.tie_word_embeddings:
            self._tie_or_clone_weights(self.encoder.embed_tokens, self.shared)
            self._tie_or_clone_weights(self.decoder.embed_tokens, self.shared)

    def get_input_embeddings(self):
        return self.shared

    def set_input_embeddings(self, value):
        self.shared = value
        self.encoder.embed_tokens = self.shared
        self.decoder.embed_tokens = self.shared

    def get_encoder(self):
        return self.encoder

    def get_decoder(self):
        return self.decoder

    @add_start_docstrings_to_model_forward(BART_INPUTS_DOCSTRING)
    @add_code_sample_docstrings(
        checkpoint=_CHECKPOINT_FOR_DOC,
        output_type=Seq2SeqModelOutput,
        config_class=_CONFIG_FOR_DOC,
        expected_output=_EXPECTED_OUTPUT_SHAPE,
    )
    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        decoder_input_ids: Optional[torch.LongTensor] = None,
        decoder_attention_mask: Optional[torch.LongTensor] = None,
        decoder_position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        decoder_head_mask: Optional[torch.Tensor] = None,
        cross_attn_head_mask: Optional[torch.Tensor] = None,
        encoder_outputs: Optional[List[torch.FloatTensor]] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        decoder_inputs_embeds: Optional[torch.FloatTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, Seq2SeqModelOutput]:
        # different to other models, Bart automatically creates decoder_input_ids from
        # input_ids if no decoder_input_ids are provided
        if decoder_input_ids is None and decoder_inputs_embeds is None:
            if input_ids is None:
                raise ValueError(
                    "If no `decoder_input_ids` or `decoder_inputs_embeds` are "
                    "passed, `input_ids` cannot be `None`. Please pass either "
                    "`input_ids` or `decoder_input_ids` or `decoder_inputs_embeds`."
                )

            decoder_input_ids = shift_tokens_right(
                input_ids, self.config.pad_token_id, self.config.decoder_start_token_id
            )

        output_attentions = (
            output_attentions if output_attentions is not None else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        use_cache = use_cache if use_cache is not None else self.config.use_cache
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if encoder_outputs is None:
            encoder_outputs = self.encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                head_mask=head_mask,
                inputs_embeds=inputs_embeds,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
                return_dict=return_dict,
            )
        # If the user passed a tuple for encoder_outputs, we wrap it in a BaseModelOutput when return_dict=True
        elif return_dict and not isinstance(encoder_outputs, BaseModelOutput):
            encoder_outputs = BaseModelOutput(
                last_hidden_state=encoder_outputs[0],
                hidden_states=encoder_outputs[1] if len(encoder_outputs) > 1 else None,
                attentions=encoder_outputs[2] if len(encoder_outputs) > 2 else None,
            )

        if not (
            isinstance(encoder_outputs, BaseModelOutput) or isinstance(encoder_outputs, tuple)
        ):
            raise ValueError(
                "Inconsistent output: The output of the model encoder should be of type "
                f"`BaseModelOutput` or tuple, but is of type `{type(encoder_outputs)}`."
            )

        # decoder outputs consists of (dec_features, past_key_value, dec_hidden, dec_attn)
        decoder_outputs = self.decoder(
            input_ids=decoder_input_ids,
            attention_mask=decoder_attention_mask,
            position_ids=decoder_position_ids,
            encoder_hidden_states=encoder_outputs[0],
            encoder_attention_mask=attention_mask,
            head_mask=decoder_head_mask,
            cross_attn_head_mask=cross_attn_head_mask,
            past_key_values=past_key_values,
            inputs_embeds=decoder_inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        if not return_dict:
            return decoder_outputs + encoder_outputs

        return Seq2SeqModelOutput(
            last_hidden_state=decoder_outputs.last_hidden_state,
            past_key_values=decoder_outputs.past_key_values,
            decoder_hidden_states=decoder_outputs.hidden_states,
            decoder_attentions=decoder_outputs.attentions,
            cross_attentions=decoder_outputs.cross_attentions,
            encoder_last_hidden_state=encoder_outputs.last_hidden_state,
            encoder_hidden_states=encoder_outputs.hidden_states,
            encoder_attentions=encoder_outputs.attentions,
        )
